"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import math
import copy
import enum
import itertools
import numpy as np
from typing import *
import networkx as nx
from collections import OrderedDict
from adsg_core.graph.graph_edges import *

__all__ = ['DSGNode', 'ChoiceNode', 'SelectionChoiceNode', 'ConnectionChoiceNode', 'ConnectorNode', 'NamedNode',
           'ConnectorDegreeGroupingNode', 'DesignVariableNode', 'MetricNode', 'MetricType', 'EdgeType', 'EdgeTuple',
           'NodeExportShape', 'ADSGNode']


class NodeExportShape(enum.Enum):
    CIRCLE = 0
    ROUNDED_RECT = 1
    HEXAGON = 2


_CHOICE_COLOR = '#81D4FA'
_INP_OUT_COLOR = '#DCE775'


class DSGNode:
    """
    Base class for any node used in the DSG.

    The decision_id attribute is used for grouping nodes into selection-choices to be taken. The option_id attribute
    is used for sorting choice options (if applicable).
    """

    def __init__(self, obj_id=None, decision_id=None, option_id=None, src_decision_link_key=None,
                 tgt_decision_link_key=None, perm_decision_link_key=None, obj_ref=None):
        self._obj_id = obj_id
        self._id = None
        self.update_node_id()
        self.obj_ref = copy.deepcopy(obj_ref)

        self.decision_id = decision_id
        self.option_id = option_id
        self.src_decision_link_key = src_decision_link_key
        self.tgt_decision_link_key = tgt_decision_link_key
        self.perm_decision_link_key = perm_decision_link_key

    def update_node_id(self):
        self._id = hash(self._obj_id or id(self))

    def get_export_shape(self) -> NodeExportShape:
        return NodeExportShape.CIRCLE

    def get_export_color(self) -> str:
        return '#ffffff'

    def get_export_title(self) -> str:
        """Human-readable title for exporting"""
        return str(self)

    def __hash__(self):
        return self._id

    def __eq__(self, other):
        return self._id == hash(other)

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return str(self)

    def str_context(self):
        """Technical title for exporting"""
        return str(self)

    def copy_node(self, i=None):
        """Copy the node without reference"""
        node_copy = copy.copy(self)
        node_copy.obj_ref = copy.deepcopy(node_copy.obj_ref)
        node_copy.update_node_id()
        return node_copy


ADSGNode = DSGNode  # Backward compatibility


class NamedNode(DSGNode):
    """Generic node with a name"""

    def __init__(self, name: str, **kwargs):
        self.name = name
        super().__init__(**kwargs)

    def get_export_title(self) -> str:
        return self.name

    def __str__(self):
        return f'[{self.name}]'


EdgeTuple = Tuple[DSGNode, DSGNode, Hashable, HashableDict]


class ConnectorNode(DSGNode):
    """
    A node specifying some connection to be made. Specifies data about the number of connections possible using the
    deg_list attribute (specifying a list of allowed number of connections), or using the deg_min and deg_max
    attributes, which specify a range (inclusive). Note that deg_max can also be inf (math.inf), denoting no upper
    limit.
    """

    def __init__(self, name: str = None, deg_spec=None, deg_list=None, deg_min=1, deg_max=None, repeated_allowed=False,
                 **kwargs):
        self.name = name
        if deg_spec is not None:
            deg_list, deg_min, deg_max = self._parse_deg_spec(deg_spec)

        if deg_list is None:
            if deg_min < 0:
                raise ValueError('Min degree should be positive (given: %d)' % deg_min)
            if deg_max is None:
                deg_max = math.inf
            if deg_max < deg_min:
                raise ValueError('Max degree should be equal to or higher than min degree (%d >= %d)' %
                                 (deg_max, deg_min))
            if deg_min == deg_max and deg_min == math.inf:
                raise ValueError('Infinite should be part of a range, not of a single value')
        else:
            deg_min = None
            deg_max = None

        self.deg_list = deg_list
        self.deg_min = deg_min
        self.deg_max = deg_max
        self.repeated_allowed = repeated_allowed

        super(ConnectorNode, self).__init__(**kwargs)

    def is_valid(self, degree):
        """Check whether the given connection degree is valid for the connection constraint"""
        if self.deg_list is not None:
            return degree in self.deg_list
        return self.deg_min <= degree <= self.deg_max

    def update_deg(self, graph: nx.MultiDiGraph, existing_nodes: Set[DSGNode] = None):
        pass

    @staticmethod
    def _parse_deg_spec(deg_spec):
        """
        Parse various degree specification formats:
        - list of allowed degrees
        - "x..y": range of degrees, where y can also be "inf" for no upper ceiling
        - "*" for zero or more
        - "+" for one or more
        - "?" for zero or one
        - "req": precisely one
        - "opt": zero or one
        - int: precisely the amount

        :param deg_spec:
        :return:
        """

        if isinstance(deg_spec, (list, tuple)):  # Degree list
            return deg_spec, None, None

        if deg_spec in ('?', 'opt'):  # Zero or one
            deg_min, deg_max = 0, 1
        elif deg_spec == '*':  # Zero or more
            deg_min, deg_max = 0, math.inf
        elif deg_spec == '+':  # One or more
            deg_min, deg_max = 1, math.inf
        elif deg_spec == 'req':  # Precisely 1
            return [1], None, None
        else:
            try:
                if '..' in deg_spec:  # Range
                    deg_min, deg_max = deg_spec.split('..')
                    deg_min = int(deg_min)
                    deg_max = math.inf if deg_max in ('inf', '*') else int(deg_max)
                else:
                    return [int(deg_spec)], None, None
            except TypeError:  # A specified amount
                return [int(deg_spec)], None, None
        return None, deg_min, deg_max

    @classmethod
    def get_deg_kwargs(cls, deg_spec):
        """
        Get the degree-related port init kwargs, from various specification formats:
        - list of allowed degrees
        - "x..y": range of degrees, where y can also be "inf" for no upper ceiling
        - "*" for zero or more
        - "+" for one or more
        - "?" for zero or one
        - "req": precisely one
        - "opt": zero or one
        - int: precisely the amount

        :param deg_spec:
        :return:
        """
        deg_list, deg_min, deg_max = cls._parse_deg_spec(deg_spec)
        return {'deg_list': deg_list, 'deg_min': deg_min, 'deg_max': deg_max}

    @property
    def deg_str(self):
        if self.deg_list is not None:
            return ' @ %s' % (','.join([str(deg) for deg in self.deg_list]),)
        elif self.deg_min == self.deg_max:
            return '' if self.deg_min == 1 else (' @ %d' % self.deg_min)
        else:
            return ' @ %s..%s' % (self.deg_min, self.deg_max)

    def get_full_deg_str(self):
        if self.deg_list is not None:
            return ','.join([str(deg) for deg in self.deg_list])
        if self.deg_min == self.deg_max:
            return str(self.deg_min)
        if math.isinf(self.deg_max):
            return f'{self.deg_min}..*'
        return f'{self.deg_min}..{self.deg_max}'

    def get_export_title(self) -> str:
        return self.name

    def get_export_shape(self) -> NodeExportShape:
        return NodeExportShape.HEXAGON

    def __str__(self):
        name_str = f'[{self.name}]' if self.name is not None else ''
        return f'Conn{name_str}{self.deg_str}'


class ConnectorDegreeGroupingNode(ConnectorNode):
    """
    Groups a bunch of connectors. Used when the connection order of the connectors does not matter. This means that this
    class only looks at the aggregate number of connections accepted by the grouped connectors. The update_deg function
    should be called with the grouped nodes to update the connection degree specifications.
    """

    def __init__(self, name: str = None):
        super(ConnectorDegreeGroupingNode, self).__init__(name)

    def __str__(self):
        return 'Grp[Conn]'

    def update_deg(self, graph: nx.MultiDiGraph, existing_nodes: Set[DSGNode] = None):

        connectors: List[ConnectorNode] = \
            [in_edge[0] for in_edge in iter_in_edges(graph, self, edge_type=EdgeType.DERIVES)]
        if existing_nodes is not None:
            connectors = [node for node in connectors if node in existing_nodes]

        self.deg_list, self.deg_min, self.deg_max = self.get_combined_deg(connectors)
        self.repeated_allowed = self.get_repeated_allowed(connectors)

        # Update decision link key
        for node in connectors:  # type: DSGNode
            if node.perm_decision_link_key:
                self.perm_decision_link_key = node.perm_decision_link_key
                break

    @staticmethod
    def get_repeated_allowed(connectors: List[ConnectorNode]) -> bool:
        for connector in connectors:
            if connector.repeated_allowed:
                return True
        return False

    @staticmethod
    def get_combined_deg(connectors: List[ConnectorNode]) \
            -> Tuple[Optional[List[int]], Optional[int], Optional[Union[int, float]]]:

        # Gather specifications from connectors
        deg_lists = []
        deg_min_inf = None
        for connector in connectors:
            conn_deg_list = connector.deg_list
            conn_deg_min, conn_deg_max = connector.deg_min, connector.deg_max
            if conn_deg_list is not None:
                deg_lists.append(conn_deg_list)
            else:
                if conn_deg_max != math.inf:
                    deg_lists.append(list(range(conn_deg_min, conn_deg_max+1)))
                else:
                    if deg_min_inf is None:
                        deg_min_inf = 0
                    deg_min_inf += conn_deg_min

        # If there is an infinite upper limit, treat the allowed degrees as a list
        if deg_min_inf is not None:
            deg_min_inf += sum([min(deg_list) for deg_list in deg_lists])
            return None, deg_min_inf, math.inf

        # Get all unique connection combinations
        deg_list = sorted(list({sum(comb) for comb in itertools.product(*deg_lists)}))
        return deg_list, None, None

    def get_export_title(self) -> str:
        return 'Grp'

    def get_export_shape(self) -> NodeExportShape:
        return NodeExportShape.HEXAGON


class DesignVariableNode(DSGNode):
    """
    Node representing a design variable.
    """

    def __init__(self, name, bounds=None, options=None, idx=None, **kwargs):
        if bounds is not None:
            if len(bounds) != 2:
                raise ValueError('Bounds should be a tuple or list of length 2')
            if bounds[0] >= bounds[1]:
                raise ValueError('Upper bound should be higher than lower bound')
            if options is not None:
                raise ValueError('Either bounds or options should be provided, not both!')

        elif options is not None:
            if len(options) < 1:
                raise ValueError('Options list should contain at least one value')

        self.name = name
        self.bounds = bounds
        self.options = options
        self.idx = idx
        self.assigned_value = None  # Only for export!
        super(DesignVariableNode, self).__init__(**kwargs)

        if self.decision_id is None:
            self.decision_id = f'DV_{self.name}'

    @property
    def is_discrete(self):
        """Whether the design variable is discrete or continuous"""
        return self.options is not None

    def correct_value(self, value):
        """Ensure a value is within the bounds or represents a valid option"""
        if self.bounds is None and self.options is None:
            raise ValueError(f'Design variable bounds or options not set: {self!r}')

        bounds_fraction = .5
        if self.is_discrete:
            if value < 0:
                value = 0
            elif value >= len(self.options):
                value = len(self.options)-1
        else:
            lower, upper = self.bounds
            if value < lower:
                value = lower
            elif value > upper:
                value = upper
            bounds_fraction = (value - lower) / (upper - lower)
        return value, bounds_fraction

    def get_export_title(self) -> str:
        if self.assigned_value is not None:
            if self.is_discrete:
                values_str = f'= {self.options[self.assigned_value]}'
            else:
                values_str = f'= {self.assigned_value:.4g}'
        else:
            if self.is_discrete:
                values_str = ' ['+','.join([str(opt) for opt in self.options])+']'
            else:
                values_str = f' [{self.bounds[0]} .. {self.bounds[1]}]'
        return f'{self.name} {values_str}'

    def get_export_color(self) -> str:
        return _INP_OUT_COLOR

    def str_context(self):
        return f'DV.{self.name}.{self.idx}.{self.bounds!r}.{self.options!r}'

    def __str__(self):
        return f'DV[{self.name}]'


class MetricType(enum.Flag):
    NONE = 0
    OBJECTIVE = enum.auto()
    CONSTRAINT = enum.auto()
    OBJ_OR_CON = OBJECTIVE | CONSTRAINT


class MetricNode(DSGNode):
    """
    Node representing a performance metric. In the end, a performance metric can either be used as an
    objective or as a constraint in an optimization problem.

    If a direction (meaning which direction of the value is "better") is given, it can be used an objective. If also a
    reference is given (meaning it should be in a specified direction, but at least this value), it can also be used
    as a constraint.

    Metric type is derived automatically, however to force the type you can supply a `MetricType`:

    - `NONE`: no role in the optimization problem
    - `OBJECTIVE`: metric is an objective (provide a direction)
    - `CONSTRAINT`: metric is a constraint (provide a direction and a reference value)
    """

    def __init__(self, name, direction: int = None, ref: float = None, idx=None, type_=None):
        self.name = name
        self.idx = idx
        self.dir = direction  # -1 for min/lte, 1 for max/gte
        self.ref = ref  # Reference value for constraint
        self.type: Optional[MetricType] = type_
        self.assigned_value = None  # Only for export!
        super(MetricNode, self).__init__()

    def get_export_title(self) -> str:
        role_str = ''
        if self.dir is not None:
            if self.ref is not None:
                role_str = ' [>=' if self.dir > 0 else ' [<='
                role_str += f' {self.ref!s}]'
            else:
                role_str = ' [max]' if self.dir > 0 else ' [min]'

        if self.assigned_value is not None:
            if math.isnan(self.assigned_value):
                role_str = f' = NaN'+role_str
            else:
                role_str = f' = {self.assigned_value:.4g}'+role_str

        return self.name+role_str

    def get_export_color(self) -> str:
        return _INP_OUT_COLOR

    def get_export_shape(self) -> NodeExportShape:
        return NodeExportShape.ROUNDED_RECT

    def str_context(self):
        return f'MET.{self.name}.{self.idx}.dir{self.dir}.ref{self.ref}'

    def __str__(self):
        return 'PR[%s]' % self.name


class ChoiceNode(DSGNode):
    """
    Node representing a decision to be taken.
    """

    def __init__(self, decision_id=None, decision_sort_key=None):
        self.decision_sort_key = decision_sort_key
        super(ChoiceNode, self).__init__(decision_id=decision_id)

    def get_export_title(self) -> str:
        return self.decision_id

    def get_export_color(self) -> str:
        return _CHOICE_COLOR

    def __str__(self):
        raise NotImplementedError


class SelectionChoiceNode(ChoiceNode):
    """
    Node representing a selection: a choice with a discrete set of mutually exclusive options to
    choose from.
    """

    def __str__(self):
        return 'D[Sel: %s]' % (self.decision_id or '_')

    def str_context(self):
        return 'D[Sel: %s]' % self.decision_id


class ConnectionChoiceNode(ChoiceNode):
    """
    Node representing a connection choice: a choice about how to connect between a set of source nodes and a set of
    target nodes. Sources and targets should be `ConnectorNode` types.
    """

    def __init__(self, decision_id=None, decision_sort_key=None):
        super(ConnectionChoiceNode, self).__init__(decision_id=decision_id, decision_sort_key=decision_sort_key)

    def __str__(self):
        return 'D[Conn: %s]' % (self.decision_id or '_')

    def str_context(self):
        return 'D[Conn: %s]' % self.decision_id

    def iter_conn_edges(self, dsg, check_conditional=False):
        from adsg_core.optimization.assign_enc.matrix import NodeExistence

        matrix_gen, node_map = self._get_matrix_gen(dsg, check_conditional=check_conditional)
        agg_matrix = matrix_gen.get_agg_matrix(cache=True)[NodeExistence()]
        for i_mat in range(agg_matrix.shape[0]):
            conn_edges = matrix_gen.get_conn_idx(agg_matrix[i_mat, :, :])
            node_edges = [(node_map[0][conn_edge[0]], node_map[1][conn_edge[1]]) for conn_edge in conn_edges]
            yield node_edges

    def validate_conn_edges(self, dsg, edges: Sequence[Tuple[ConnectorNode, ConnectorNode]]) -> bool:
        matrix_gen, node_map = self._get_matrix_gen(dsg)
        src_idx_map = {node: i for i, node in enumerate(node_map[0])}
        tgt_idx_map = {node: i for i, node in enumerate(node_map[1])}

        matrix = np.zeros((len(node_map[0]), len(node_map[1])), dtype=int)
        for src, tgt in edges:
            if src not in src_idx_map or tgt not in tgt_idx_map:
                return False
            matrix[src_idx_map[src], tgt_idx_map[tgt]] += 1

        return matrix_gen.validate_matrix(matrix)

    def _get_matrix_gen(self, dsg, check_conditional=False):
        from adsg_core.optimization.assign_enc.matrix import AggregateAssignmentMatrixGenerator

        settings, node_map = self._get_assign_nodes(dsg, check_conditional=check_conditional)
        matrix_gen = AggregateAssignmentMatrixGenerator(settings)
        return matrix_gen, node_map

    def get_assignment_encoding_args(self, dsg, hierarchy_analyzer=None):
        from adsg_core.optimization.assign_enc.matrix import NodeExistencePatterns, NodeExistence, MatrixGenSettings
        from adsg_core.optimization.hierarchy import HierarchyAnalyzer

        if hierarchy_analyzer is None:
            hierarchy_analyzer = HierarchyAnalyzer(dsg)

        # Create connection nodes
        settings, node_map = self._get_assign_nodes(dsg)

        # Get existence patterns
        flat_idx_map = {}
        conn_nodes = list(node_map[0])+list(node_map[1])
        derivation_nodes = self.get_conn_node_derivations(dsg.graph, conn_nodes)
        all_conn_nodes = []
        for conn_node, deriving_conn_nodes in derivation_nodes.items():
            flat_idx_map[conn_node] = len(all_conn_nodes)
            all_conn_nodes.append(conn_node)

            for deriving_node in deriving_conn_nodes:
                flat_idx_map[deriving_node] = len(all_conn_nodes)
                all_conn_nodes.append(deriving_node)

        def _exist_process(assign_nodes, exists, n_conn_override, sub_node_map, n_max):
            for ii in range(len(assign_nodes)):
                conn_node_ = sub_node_map[ii]

                # Check connector node existence
                if not existence_mask[flat_idx_map[conn_node_]]:
                    exists[ii] = False
                    continue

                # Check connection number override
                if isinstance(conn_node_, ConnectorDegreeGroupingNode) and len(derivation_nodes[conn_node_]) > 0:
                    existing_connectors = [der_conn_node for der_conn_node in derivation_nodes[conn_node_]
                                           if existence_mask[flat_idx_map[der_conn_node]]]
                    deg_list, deg_min, deg_max = conn_node_.get_combined_deg(existing_connectors)
                    if deg_list is None:
                        if deg_max == math.inf:
                            deg_max = n_max[ii]
                        deg_list = list(range(deg_min, (deg_max or deg_min)+1))
                    n_conn_override[ii] = deg_list

        # Get possible connector node existences, if not available, generate the cartesian product
        map_by_existence_mask = False
        existence_masks = hierarchy_analyzer.get_nodes_existence(all_conn_nodes)
        if existence_masks is not None:
            unique_existence_masks, existence_map = np.unique(existence_masks, axis=0, return_inverse=True)

        else:
            map_by_existence_mask = True
            unique_existence_masks = np.array(list(itertools.product(
                *[[False, True] for _ in range(len(all_conn_nodes))])))
            existence_map = np.arange(0, len(unique_existence_masks))

        max_conn_matrix = settings.get_max_conn_matrix()

        patterns = []
        pattern_idx_map = {}
        src_nodes, tgt_nodes = settings.src, settings.tgt
        for i_exist, existence_mask in enumerate(unique_existence_masks):
            src_exists, tgt_exists = [True for _ in range(len(src_nodes))], [True for _ in range(len(tgt_nodes))]
            src_n_conn_override, tgt_n_conn_override = {}, {}

            _exist_process(src_nodes, src_exists, src_n_conn_override, node_map[0], np.sum(max_conn_matrix, axis=1))
            _exist_process(tgt_nodes, tgt_exists, tgt_n_conn_override, node_map[1], np.sum(max_conn_matrix, axis=0))

            existence = NodeExistence(
                src_exists=src_exists, tgt_exists=tgt_exists,
                src_n_conn_override=src_n_conn_override, tgt_n_conn_override=tgt_n_conn_override,
            )

            # Prevent duplicates
            if existence in pattern_idx_map:
                existence_map[existence_map == i_exist] = pattern_idx_map[existence]
                continue

            pattern_idx_map[existence] = i_pattern = len(patterns)
            existence_map[existence_map == i_exist] = i_pattern
            patterns.append(existence)

        existence_patterns = NodeExistencePatterns(patterns=patterns)
        settings = MatrixGenSettings(settings.src, settings.tgt, settings.excluded, existence=existence_patterns)

        # Enable identification by node existence rather than combination idx
        if map_by_existence_mask:
            existence_map = {tuple(unique_existence_masks[i_mask]): i_pattern
                             for i_mask, i_pattern in enumerate(existence_map)}

        return settings, node_map, existence_map, all_conn_nodes

    def _get_assign_nodes(self, dsg, check_conditional=False):
        from adsg_core.optimization.assign_enc.matrix import Node as AssignNode, MatrixGenSettings

        graph = dsg.graph
        src_nodes = self.get_src_nodes(graph)
        tgt_nodes = self.get_tgt_nodes(graph)

        for conn_nodes in [src_nodes, tgt_nodes]:
            for conn_node in conn_nodes:
                if isinstance(conn_node, ConnectorDegreeGroupingNode):
                    conn_node.update_deg(dsg.graph)

        # Check whether any of the nodes exist conditionally
        def _is_conditional(connector_node):
            if dsg.has_conditional_existence(connector_node):
                return True

            # Check ancestor port nodes if this is a port grouping node
            if isinstance(connector_node, ConnectorDegreeGroupingNode):
                for prev_node in graph.predecessors(connector_node):
                    if dsg.has_conditional_existence(prev_node):
                        return True

            return False

        src_conditional = set()
        tgt_conditional = set()
        if check_conditional:
            src_conditional = {node for node in src_nodes if _is_conditional(node)}
            tgt_conditional = {node for node in tgt_nodes if _is_conditional(node)}

        src_obj_map = OrderedDict([(node, self.to_assign_node(node, is_conditional=node in src_conditional))
                                   for node in src_nodes])
        tgt_obj_map = OrderedDict([(node, self.to_assign_node(node, is_conditional=node in tgt_conditional))
                                   for node in tgt_nodes])
        node_map = (src_nodes, tgt_nodes)

        src_objs: List[AssignNode] = list(src_obj_map.values())
        tgt_objs: List[AssignNode] = list(tgt_obj_map.values())

        # Get excluded edges
        excluded = []
        for edge in self.get_excluded_edges(graph):
            if edge[0] not in src_obj_map or edge[1] not in tgt_obj_map:
                continue
            excluded.append((src_obj_map[edge[0]], tgt_obj_map[edge[1]]))

        settings = MatrixGenSettings(src_objs, tgt_objs, excluded)
        return settings, node_map

    def get_src_nodes(self, graph) -> List[ConnectorNode]:
        return self.get_sorted_connector_nodes([node for node in graph.predecessors(self)])

    def get_tgt_nodes(self, graph) -> List[ConnectorNode]:
        return self.get_sorted_connector_nodes([node for node in graph.successors(self)])

    @staticmethod
    def get_conn_node_derivations(graph, nodes: List[ConnectorNode]) \
            -> Dict[ConnectorNode, List[ConnectorNode]]:
        node_derivations = {}
        for node in nodes:
            deriving_nodes = []
            if isinstance(node, ConnectorDegreeGroupingNode):
                for prev_node in graph.predecessors(node):
                    if isinstance(prev_node, ConnectorNode):
                        deriving_nodes.append(prev_node)
            node_derivations[node] = deriving_nodes
        return node_derivations

    @staticmethod
    def get_sorted_connector_nodes(nodes):
        return sorted([node for node in nodes if isinstance(node, ConnectorNode)],
                      key=lambda n: (str(n.option_id or ''), str(n.decision_id or '')))

    def get_excluded_edges(self, graph):
        excluded = []
        for node in self.get_src_nodes(graph):
            excluded += [edge for edge in iter_out_edges(graph, node, edge_type=EdgeType.EXCLUDES)]
        return excluded

    def get_deriving_edges(self, graph):
        tgt_nodes = self.get_tgt_nodes(graph)
        deriving_edges = []
        for node in self.get_src_nodes(graph):
            for edge in iter_out_edges(graph, node):
                if edge[1] in tgt_nodes and get_edge_type(edge) == EdgeType.DERIVES:
                    deriving_edges.append(edge)
        return deriving_edges

    @staticmethod
    def to_assign_node(connector_node: ConnectorNode, is_conditional=False):
        from adsg_core.optimization.assign_enc.matrix import Node as AssignNode

        deg_list = connector_node.deg_list
        deg_min = connector_node.deg_min
        deg_max = connector_node.deg_max
        if deg_max is None:
            deg_max = deg_min

        # If this node exists conditionally, simply add the option to accept 0 connections
        if is_conditional:
            if deg_list is not None:
                deg_list = [0]+list(deg_list)
            else:
                deg_min = 0

        return AssignNode(
            nr_conn_list=deg_list,
            min_conn=deg_min,
            max_conn=deg_max,
            repeated_allowed=connector_node.repeated_allowed,
        )

    def get_export_shape(self) -> NodeExportShape:
        return NodeExportShape.HEXAGON



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.encoding import *

__all__ = ['AutoModImputer']


class AutoModImputer(EagerImputer):
    """
    Imputes design vectors by moving from the start or end to change design variables if they are too restrictive:
    For each design variable (from start or end):
    - Create a new design vector only fixing the values up to that design vector
    - Filter out design vectors matching this partial vector
    - If none were found: try to assign different values to the design variable (skipping the current value)
      - If a value is found that matches 1 or more design vectors, stop the search for a different value
    - If exactly one design vector was found to comply: return this as the imputed design vector
    - If more than one design vectors were found: move to the next design variable
    """

    def __init__(self, reverse=False):
        super().__init__()
        self.reverse = reverse

    def impute(self, vector: DesignVector, existence: NodeExistence, matrix_mask: MatrixSelectMask) -> Tuple[DesignVector, np.ndarray]:
        if len(vector) == 0:
            return vector, np.zeros((0, 0), dtype=int)

        mask = matrix_mask
        partial_vector: PartialDesignVector = [None for _ in range(len(vector))]
        for i_dv in (reversed(range(len(vector))) if self.reverse else range(len(vector))):
            dv_partial_vector = partial_vector.copy()
            dv_partial_vector[i_dv] = vector[i_dv]

            # Find design vectors that match the new partial design vector
            dv_mask = self._filter_design_vectors(dv_partial_vector, existence) & mask
            i_mask, = np.where(dv_mask)

            # If we have found no valid design vectors for this partial vector, modify the value until we have found one
            if len(i_mask) == 0:
                for dv_value in range(self._design_vars[i_dv].n_opts):
                    # Skip current value
                    if dv_value == vector[i_dv]:
                        continue

                    # Find design vectors that match the modified partial design vector
                    dv_partial_vector[i_dv] = dv_value
                    dv_mask = self._filter_design_vectors(dv_partial_vector, existence) & mask
                    i_mask, = np.where(dv_mask)

                    if len(i_mask) > 0:
                        break

                if len(i_mask) == 0:
                    raise RuntimeError('Cannot find any valid design vectors!')

            # If we have found exactly one vector, we can stop searching and return this one
            if len(i_mask) == 1:
                return self._return_imputation(i_mask[0], existence)

            # Otherwise, continue with the next design variable
            else:
                mask = dv_mask
                partial_vector = dv_partial_vector

        # Only happens if there are duplicate design vectors
        raise RuntimeError('Multiple possible design vectors found! Check if there are any duplicates')

    def __repr__(self):
        return f'{self.__class__.__name__}(reverse={self.reverse})'

    def __str__(self):
        return f'Auto Mod Imp{" Rev" if self.reverse else ""}'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import itertools
import numpy as np
from typing import *
from enum import Enum
from dataclasses import dataclass
from adsg_core.graph.adsg_nodes import *

__all__ = ['CDVNode', 'ChoiceConstraint', 'ChoiceConstraintType', 'get_constraint_removed_options',
           'get_constraint_pre_removed_options', 'get_valid_idx_combinations', 'count_n_combinations_max',
           'CCT_EXPORT_LABEL']

CDVNode = Union[ChoiceNode, DesignVariableNode]


class ChoiceConstraintType(Enum):
    LINKED = 1  # To make all choices have the same option index             --> AA, BB, CC
    PERMUTATION = 2  # To make the choices be permutations of option indices --> AB, BA, AC, CA, BC, CB
    UNORDERED = 3  # To have all option index combinations without ordering  --> AA, AB, AC, BB, BC, CC
    UNORDERED_NOREPL = 4  # Same but also without replacement                --> AB, AC, BC


CCT_EXPORT_LABEL = {
    ChoiceConstraintType.LINKED: '=',
    ChoiceConstraintType.PERMUTATION: '≠',
    ChoiceConstraintType.UNORDERED: '≥',
    ChoiceConstraintType.UNORDERED_NOREPL: '>',
}


@dataclass(frozen=True)
class ChoiceConstraint:
    type: ChoiceConstraintType
    nodes: List[CDVNode]
    options: List[list] = None

    def __hash__(self):
        return id(self)


def get_constraint_removed_options(choice_constraint: ChoiceConstraint, i_taken_choice: int, i_chosen_option: int) \
        -> List[Tuple[ChoiceNode, list]]:
    """
    Implements the option-removal logic, regardless of which choice is taken first.

    Consider two choices (D1, D2) with three options each (A, B, C).
    For a LINKED choice constraint: remove all options with different indices
    D1 = A --> remove D2 BC --> D2 = A; D1+D2 = AA
    D2 = B --> remove D1 AC --> D2 = B; D1+D2 = BB
    all D1+D2 options: AA, BB, CC

    For a PERMUTATION constraint: remove all options with the same index
    D1 = A --> remove D2 A --> D2 = B or C; D1+D2 = AB, AC
    D1 = B --> remove D2 B --> D2 = A or C; D1+D2 = BA, BC
    all D1+D2 options: AB, BA, AC, CA, BC, CB

    For an UNORDERED constraint: remove all subsequent options with lower indices (and vice-versa)
    D1 = B --> remove D2 A --> D2 = B or C; D1+D2 = BB, BC
    D2 = B --> remove D1 C --> D1 = A or B; D1+D2 = AB, BB
    all D1+D2 options: AA, AB, AC, BB, BC, CC

    For an UNORDERED_NOREPL constraint: remove all subsequent options with lower or same indices (and vice-versa)
    D1 = B --> remove D2 A,B --> D2 = C; D1+D2 = BC
    D2 = B --> remove D1 B,C --> D1 = A; D1+D2 = AB
    all D1+D2 options: AB, AC, BC --> observe that D1=C or D2=A lead to infeasible architectures
                                      (prevented using `get_constraint_pre_removed_options`)
    """
    if choice_constraint.options is None:
        return []

    removed_choice_opts = []
    for i, choice_node in enumerate(choice_constraint.nodes):
        if i == i_taken_choice:
            continue
        options = choice_constraint.options[i]
        enough_options = len(options)-1 >= i_chosen_option

        removed_opts = []
        if choice_constraint.type == ChoiceConstraintType.LINKED:
            # For linked constraints, we simply remove all options with different indices from the other choices
            # If a choice does not have as many options, we remove all but the last
            if enough_options:
                removed_opts = [opt for j, opt in enumerate(options) if j != i_chosen_option]
            else:
                removed_opts = options[:-1]

        elif choice_constraint.type == ChoiceConstraintType.PERMUTATION:
            # For permutation constraints, we remove all options with the same index from the other choices
            if enough_options:
                removed_opts = [options[i_chosen_option]]

        elif choice_constraint.type == ChoiceConstraintType.UNORDERED:
            # For unordered combination constraints, the goal is that subsequent choices cannot take options with
            # lower indices (and therefore preceding choices cannot have options with higher indices)
            if i < i_taken_choice:
                removed_opts = options[i_chosen_option+1:]  # Preceding cannot have higher-index options
            else:
                removed_opts = options[:i_chosen_option]  # Subsequent cannot have lower-index options

        elif choice_constraint.type == ChoiceConstraintType.UNORDERED_NOREPL:
            # Unordered non-replacing combination constraints behave almost the same as unordered, however here also
            # taking the same index (which would be replacement) is not allowed
            if i < i_taken_choice:
                removed_opts = options[i_chosen_option:]  # Preceding cannot have higher or the same index
            else:
                removed_opts = options[:i_chosen_option+1]  # Subsequent cannot have lower or the same index

        else:
            raise RuntimeError(f'Unsupported constraint type: {choice_constraint.type}')
        if len(removed_opts) > 0:
            removed_choice_opts.append((choice_node, removed_opts))
    return removed_choice_opts


def get_constraint_pre_removed_options(choice_constraint: ChoiceConstraint, permanent_nodes: Set[DSGNode]) \
        -> List[Tuple[ChoiceNode, list]]:
    """
    Get options that can never be selected in order to prevent constraints that are impossible to satisfy.
    """
    if choice_constraint.options is None:
        return []

    # For permutations, if there are more choices that the max nr of options, there is no way to make a permutation
    if choice_constraint.type == ChoiceConstraintType.PERMUTATION:
        n_dec = len(choice_constraint.nodes)
        n_opt_max = max([len(options) for options in choice_constraint.options])

        # Remove all options if there are more choices than the max nr of options
        if n_dec > n_opt_max:
            return [(dec_node, choice_constraint.options[i]) for i, dec_node in enumerate(choice_constraint.nodes)]

    # For unordered non-replacing choices, remove options to ensure infeasible combinations never get selected
    # We can only apply this if all choices are permanent, otherwise we can't be sure
    if choice_constraint.type == ChoiceConstraintType.UNORDERED_NOREPL \
            and all(node in permanent_nodes for node in choice_constraint.nodes):
        pre_removed_opts = []
        n_dec = len(choice_constraint.nodes)
        for i_dec, dec_node in enumerate(choice_constraint.nodes):
            i_start = i_dec
            n_dec_after = n_dec-(i_dec+1)
            i_end = len(choice_constraint.options[i_dec]) - n_dec_after
            removed_options = [opt for i_opt, opt in enumerate(choice_constraint.options[i_dec])
                               if i_opt < i_start or i_opt >= i_end]
            pre_removed_opts.append((dec_node, removed_options))

        return pre_removed_opts

    return []


def get_valid_idx_combinations(idx_comb: np.ndarray, constraint_type: ChoiceConstraintType,
                               is_all_permanent=False) -> np.ndarray:
    """
    Given a matrix with n_comb x n_choices index combinations, return which combinations are valid.
    Inactive values (-1) are ignored.

    For example:
    [[0, 0],
     [0, 1],
     [1, 0],
     [1, 1]]

    for LINKED type, returns: [0, 3]
    for PERMUTATION type, returns: [1, 2]
    for UNORDERED type, returns: [0, 1, 3]
    for UNORDERED_NOREPL type: [1]

    The is_all_permanent defines whether all involved choices are permanent.
    """

    # If there is only one column, there is no constraint to be applied so all are valid
    if idx_comb.shape[1] <= 1:
        return np.arange(idx_comb.shape[0])

    valid_mask = np.ones((idx_comb.shape[0],), dtype=bool)

    def _iter_rows(is_valid):
        for i_row, row in enumerate(idx_comb):
            active_row = row[row != -1]
            if len(active_row) > 1:
                valid_mask[i_row] = is_valid(active_row)

    if constraint_type == ChoiceConstraintType.LINKED:
        _iter_rows(lambda v: np.all(v[1:] == v[0]))

    elif constraint_type == ChoiceConstraintType.PERMUTATION:
        for i, j in itertools.combinations(range(idx_comb.shape[1]), 2):
            valid_mask_i = (idx_comb[:, i] != idx_comb[:, j]) | (idx_comb[:, i] == -1) | (idx_comb[:, j] == -1)
            valid_mask &= valid_mask_i

    elif constraint_type == ChoiceConstraintType.UNORDERED or \
            (is_all_permanent and constraint_type == ChoiceConstraintType.UNORDERED_NOREPL):

        def _check_gte(row):
            for i_value in range(1, len(row)):
                if row[i_value] < row[i_value-1]:
                    return False
            return True

        _iter_rows(_check_gte)

    elif not is_all_permanent and constraint_type == ChoiceConstraintType.UNORDERED_NOREPL:

        def _check_gt(row):
            for i_value in range(1, len(row)):
                if row[i_value] <= row[i_value-1]:
                    return False
            return True

        _iter_rows(_check_gt)

    else:
        raise ValueError(f'Unknown choice constraint: {constraint_type}')

    return np.where(valid_mask)[0]


def count_n_combinations_max(choice_constraint: ChoiceConstraint, is_all_permanent=False) -> int:
    # Not implemented for continuous or connection choices
    if choice_constraint.options is None or not isinstance(choice_constraint.nodes[0], SelectionChoiceNode):
        return 0

    # Fixes counting for unordered non-replacing in case it is not known which nodes actually exist at a given time
    if choice_constraint.type == ChoiceConstraintType.UNORDERED_NOREPL and not is_all_permanent:
        is_all_permanent = True

    all_idx_comb = np.array(list(itertools.product(
        *[range(len(opt_nodes)) for opt_nodes in choice_constraint.options])))
    return len(get_valid_idx_combinations(all_idx_comb, choice_constraint.type, is_all_permanent=is_all_permanent))



import itertools
import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.encoding import *

__all__ = ['DeltaImputer']


class DeltaImputer(EagerImputer):
    """Imputes by modifying the design vector (gradually increasing the amount of mods) until a valid one is found."""

    n_max_tries = 10000

    def __init__(self):
        super().__init__()
        self._dv_idx_map = {}

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self._dv_idx_map = {existence: {tuple(dv): i for i, dv in enumerate(dvs)}
                            for existence, dvs in self._design_vectors_zero.items()}

    def impute(self, vector: DesignVector, existence: NodeExistence, matrix_mask: MatrixSelectMask) -> Tuple[DesignVector, np.ndarray]:
        design_vectors = self._get_design_vectors(existence)[matrix_mask, :]
        if len(vector) == 0 or len(design_vectors) == 0:
            return vector, np.zeros((0, 0), dtype=int)

        # Determine delta values to try out
        def _sort_by_dist(des_var_delta):
            i_sorted = np.argsort(np.abs(des_var_delta))
            return des_var_delta[i_sorted]

        delta_values = [_sort_by_dist(np.arange(dv.n_opts)-vector[i]) for i, dv in enumerate(self._design_vars)]

        dv_map = self._dv_idx_map[existence]
        n_tries, n_tries_max = 0, self.n_max_tries
        first = True
        for dv_delta in itertools.product(*delta_values):
            if first:  # Skip first as this one has a delta of 0
                first = False
                continue
            dv = tuple(vector+np.array(dv_delta))

            # Check if design vector exists
            if dv in dv_map:
                idx = dv_map[dv]
                if matrix_mask[idx]:
                    return self._return_imputation(idx, existence)

            # Limit the amount of tries
            n_tries += 1
            if n_tries > n_tries_max:
                invalid_matrix = -np.ones(self._get_matrix(existence).shape[1:], dtype=int)
                return vector, invalid_matrix

        return vector, np.zeros((0, 0), dtype=int)

    @staticmethod
    def _calc_dist_euclidean(elements: np.ndarray, target: np.ndarray) -> np.ndarray:
        return np.sqrt(np.sum((elements-target)**2, axis=1))

    @staticmethod
    def _calc_dist_manhattan(elements: np.ndarray, target: np.ndarray) -> np.ndarray:
        return np.sum(np.abs(elements-target), axis=1)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Delta Imp'



import itertools
import numpy as np
from typing import Optional
from adsg_core.optimization.assign_enc.encoding import *
from collections import defaultdict

__all__ = ['GroupedEncoder', 'GroupByIndexEncoder']


class GroupedEncoder(EagerEncoder):
    """
    Base class for an encoder that recursively separates the matrix based on some grouping criteria, until all
    matrices are associated to one unique design vector.

    By default, grouping values are normalized within sub-groups, leading to much less value options per design
    variable. To disable, set normalize_within_group=False.
    """

    def __init__(self, *args, normalize_within_group=True, **kwargs):
        self.normalize_within_group = normalize_within_group
        super().__init__(*args, **kwargs)

    def _encode(self, matrix: np.ndarray) -> np.ndarray:
        return self.group_by_values(
            self._get_grouping_values(matrix), normalize_within_group=self.normalize_within_group,
            ordinal_base=self._get_ordinal_conv_base())

    def _get_ordinal_conv_base(self) -> Optional[int]:
        pass

    @classmethod
    def group_by_values(cls, group_by_values: np.ndarray, normalize_within_group=True,
                        ordinal_base: int = None, n_declared_start=None) -> np.ndarray:
        """
        Get design vectors that uniquely map to different value combinations. Example:
        [[1 2],      [[0 0],
         [1 3],  -->  [0 1],
         [2 2],       [1 -1],
         [3 2],       [2 -1],  # -1 means inactive

        Optionally normalize only after determining all groups.
        Optionally convert the grouped design vector values to some other base,
        e.g. for base 2: [[1, 2, 3, 4]].T --> [[0, 0], [0, 1], [1, 0], [1, 1]]"""
        design_vectors = np.empty(group_by_values.shape, dtype=int)
        row_mask_list = [np.arange(group_by_values.shape[0])]

        detect_imp_ratio = cls._early_detect_high_imp_ratio
        n_valid = group_by_values.shape[0]
        if n_valid == 0:
            return np.zeros((0, 0), dtype=int)

        n_declared_total = n_declared_start if n_declared_start is not None else 1
        if detect_imp_ratio is not None and (n_declared_total/n_valid) >= detect_imp_ratio:
            raise DetectedHighImpRatio(cls, n_declared_total/n_valid)

        # Loop over columns
        col_is_active = np.ones((group_by_values.shape[1],), dtype=bool)
        for i_col in range(group_by_values.shape[1]):

            # Loop over current sub-divisions
            next_row_masks = []
            grouping_needed = False
            all_inactive = True
            col_values = set()
            for row_mask in row_mask_list:

                # Check if there are multiple design vectors in this subdivision
                group_by_values_sub = group_by_values[row_mask, i_col]
                if len(group_by_values_sub) == 1:
                    # All subsequent columns of this design vector will also be inactive
                    design_vectors[row_mask, i_col:] = X_INACTIVE_VALUE
                    continue

                # Check if there are multiple unique values in this sub-division
                value_dict = defaultdict(list)
                for value_idx, value in enumerate(group_by_values_sub):
                    value_dict[value].append(value_idx)
                unique_values = sorted(list(value_dict.keys()))

                if unique_values[0] < 0:
                    raise ValueError('Values to group by should not contain negative values!')

                if len(unique_values) == 1:
                    next_row_masks.append(row_mask)
                    design_vectors[row_mask, i_col] = X_INACTIVE_VALUE
                    grouping_needed = True
                    continue
                all_inactive = False

                # Set design vector values and get next indices for the subset of the different values
                for value_idx, value in enumerate(unique_values):
                    next_row_mask = row_mask[value_dict[value]]

                    assigned_value = value_idx if normalize_within_group else value
                    design_vectors[next_row_mask, i_col] = assigned_value
                    col_values.add(assigned_value)

                    if len(next_row_mask) == 1:
                        design_vectors[next_row_mask, i_col+1:] = X_INACTIVE_VALUE
                    else:
                        grouping_needed = True
                        next_row_masks.append(next_row_mask)

            # Detect high imputation ratios early
            if detect_imp_ratio is not None:
                # Extend the total declared values (which will correspond to the nr of options for the design vars)
                n_declared_total *= max(1, len(col_values))

                # Calculate minimum imputation ratio that can be expected: the imputation ratio can only become higher
                # by adding additional columns, so it is indeed possible to calculate the lower bound
                min_imp_ratio = n_declared_total/n_valid

                # If grouping is still needed, at least one additional design variable with minimum 2 options will be
                # added, so the lower bound can be multiplied by 2
                if grouping_needed:
                    min_imp_ratio *= 2

                if min_imp_ratio >= detect_imp_ratio:
                    raise DetectedHighImpRatio(cls, min_imp_ratio)

            # Set inactive flag
            if all_inactive:
                col_is_active[i_col] = False

            # Stop grouping if not needed anymore
            if not grouping_needed:
                design_vectors = design_vectors[:, :i_col+1]
                break

            row_mask_list = next_row_masks

        # Normalize design vectors
        if not normalize_within_group:
            design_vectors = cls.normalize_design_vectors(design_vectors)

        # Remove columns where there are no alternatives
        design_vectors = design_vectors[:, col_is_active[:design_vectors.shape[1]]]

        # Convert to other base
        if ordinal_base is not None:
            design_vectors = cls.convert_to_base(design_vectors, ordinal_base)
            has_alternatives = np.any(design_vectors > 0, axis=0)
            design_vectors = design_vectors[:, has_alternatives]

        return design_vectors

    @classmethod
    def convert_to_base(cls, values: np.ndarray, base: int = 2, n_declared_start=None) -> np.ndarray:
        """Convert the ordinal-encoded values to another base"""
        if base < 2 or base > 9:
            raise ValueError('Base should be between 2 and 9')
        if values.shape[1] == 0:
            return values

        detect_imp_ratio = cls._early_detect_high_imp_ratio
        n_valid = values.shape[0]
        if n_valid == 0:
            return np.zeros((0, 0), dtype=int)

        n_declared_total = n_declared_start if n_declared_start is not None else 1
        if detect_imp_ratio is not None and (n_declared_total/n_valid) >= detect_imp_ratio:
            raise DetectedHighImpRatio(cls, n_declared_total/n_valid)

        columns = []
        for i_col in range(values.shape[1]):
            # Get unique (active) values
            inactive_idx = []
            value_dict = defaultdict(list)
            for value_idx, value in enumerate(values[:, i_col]):
                if value == X_INACTIVE_VALUE:
                    inactive_idx.append(value_idx)
                    continue
                value_dict[value].append(value_idx)
            unique = sorted(list(value_dict.keys()))

            # Determine how many columns we need (i.e. how many digits does the largest number have)
            n_converted = len(np.base_repr(len(value_dict)-1))
            col_converted = np.zeros((values.shape[0], n_converted), dtype=int)
            col_converted[inactive_idx, :] = X_INACTIVE_VALUE

            # Convert and set values
            for i_value, value in enumerate(unique):
                base_converted = [
                    int(char) for char in (np.base_repr(i_value, base=base) if base != 2 else np.binary_repr(i_value))]
                col_converted[value_dict[value], -len(base_converted):] = base_converted

            # Early detect high imputation ratio
            if detect_imp_ratio is not None:
                n_declared_total *= np.prod(np.max(col_converted, axis=0)+1, dtype=float)
                min_imp_ratio = n_declared_total/n_valid
                if i_col < values.shape[1]-1:
                    min_imp_ratio *= 2

                if min_imp_ratio >= detect_imp_ratio:
                    raise DetectedHighImpRatio(cls, min_imp_ratio)

            columns.append(col_converted)
        return np.column_stack(columns)

    def _get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        """Get a n_mat x n_dv matrix with values to sub-divide design variables by"""
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class GroupByIndexEncoder(GroupedEncoder):
    """Grouping encoder that recursively separates the remaining matrices in n groups."""

    def __init__(self, *args, n_groups=2, **kwargs):
        self.n_groups = max(n_groups, 2)
        super().__init__(*args, **kwargs)

    def _get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        n_groups = self.n_groups
        n_var = int(np.ceil(np.log2(matrix.shape[0])))
        design_vectors = np.array(list(itertools.product(*[range(n_groups) for _ in range(n_var)]))[:matrix.shape[0]])
        for i_var in reversed(range(1, n_var)):
            design_vectors[:, i_var:] += np.array([n_groups*design_vectors[:, i_var-1]]).T
        return design_vectors

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r}, n_groups={self.n_groups})'

    def __str__(self):
        return f'Group By Index ({self.n_groups} groups) + {self._imputer!s}'



import numpy as np
from typing import Optional
from adsg_core.optimization.assign_enc.eager.encodings.grouped_base import *

__all__ = ['ElementGroupedEncoder', 'ConnIdxGroupedEncoder']


class ElementGroupedEncoder(GroupedEncoder):
    """Group by the value of each design variable"""

    def _get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        return self.flatten_matrix(matrix)

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r})'

    def __str__(self):
        normalize_str = ' Norm Grp' if self.normalize_within_group else ''
        return f'Group By Element{normalize_str} + {self._imputer!s}'


class ConnIdxGroupedEncoder(GroupedEncoder):
    """Group by the connection indices (positions); should work well for permutations"""

    def __init__(self, *args, by_src=True, binary=False, **kwargs):
        self.by_src = by_src
        self.normalize_within_group = True
        self.binary = binary
        super().__init__(*args, **kwargs)

    def _get_ordinal_conv_base(self) -> Optional[int]:
        if self.binary:
            return 2

    def _get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        if matrix.shape[0] == 0:
            return np.empty((0, 0), dtype=int)

        self.normalize_within_group = True
        conn_idx_values = []
        for i_from in range(matrix.shape[1] if self.by_src else matrix.shape[2]):
            from_matrix = matrix[:, i_from, :] if self.by_src else matrix[:, :, i_from]
            conn_idx = self.get_conn_indices(from_matrix)
            if conn_idx is not None:
                conn_idx_values.append(conn_idx)

        if len(conn_idx_values) == 0:
            return np.empty((matrix.shape[0], 0), dtype=int)
        return np.column_stack(conn_idx_values)

    @classmethod
    def get_conn_indices(cls, conn_matrix: np.ndarray) -> Optional[np.ndarray]:
        n_conn_max = np.max(np.sum(conn_matrix, axis=1))
        if n_conn_max == 0:
            return
        conn_idx = -np.ones((conn_matrix.shape[0], n_conn_max), dtype=int)
        for i, row in enumerate(conn_matrix):
            cls._set_conn_indices(row, conn_idx[i, :])
        return conn_idx+1

    @staticmethod
    def _set_conn_indices(conn_arr: np.ndarray, tgt: np.ndarray):
        offset = 0
        for i, n in enumerate(conn_arr):
            if n > 0:
                tgt[offset:offset+n] = i
                offset += n

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r}, by_src={self.by_src}, binary={self.binary})'

    def __str__(self):
        bin_str = ' Bin' if self.binary else ''
        return f'Group By {"Src" if self.by_src else "Tgt"} Conn Idx{bin_str} + {self._imputer!s}'



import gc
import ctypes
import threading
import multiprocessing.pool

try:
    import gevent
    import gevent.monkey
except ImportError:
    gevent = None

__all__ = ['run_timeout']


def run_timeout(seconds: float, func, *args, **kwargs):

    # Use native thread pool of gevent, as patched threads behave like cooperative greenlets,
    # which block if not explicitly giving control
    if gevent is not None and gevent.monkey.is_module_patched('threading'):
        raise RuntimeError('Time limiter not compatible with monkey-patched gevent threading module!')

    def _inner_run():
        with multiprocessing.pool.ThreadPool(processes=1) as pool:
            thread = pool.apply(lambda: threading.current_thread())

            try:
                return pool.apply_async(func, args, kwargs).get(timeout=seconds)
            except multiprocessing.TimeoutError:
                pass

        if thread.is_alive():
            ctypes.pythonapi.PyThreadState_SetAsyncExc(
                ctypes.c_long(thread.ident), ctypes.py_object(KeyboardInterrupt()))
            thread.join()
        raise TimeoutError

    # This call flow ensure that the memory of the "killed" thread is cleared
    try:
        return _inner_run()
    except TimeoutError:
        pass
    gc.collect()
    raise TimeoutError



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.encoding import *

__all__ = ['FirstImputer']


class FirstImputer(EagerImputer):
    """Imputer that simply chooses the first possible matrix."""

    def impute(self, vector: DesignVector, existence: NodeExistence, matrix_mask: MatrixSelectMask) -> Tuple[DesignVector, np.ndarray]:
        i_where_ok = np.where(matrix_mask)[0]
        if len(i_where_ok) == 0:
            return vector, np.zeros((0, 0), dtype=int)
        i_mat = i_where_ok[0]
        return self._return_imputation(i_mat, existence)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'First Imp'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
from typing import *
import networkx as nx
from adsg_core.graph.export import *
from adsg_core.graph.traversal import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *
from adsg_core.graph.choices import *
from adsg_core.graph.incompatibility import *
from adsg_core.graph.influence_matrix import *
from adsg_core.graph.choice_constraints import *

__all__ = ['DSG', 'EdgeType', 'CDVNode', 'ChoiceConstraint', 'ChoiceConstraintType', 'DSGType', 'ADSG', 'ADSGType']


class DSG:
    """
    The Design Space Graph (DSG). Represent the architecture design space, its choices, and contains
    functions for manipulating the graph.

    If you actually want to use the DSG directly, you can use the `BasicDSG` class in `adsg_basic`.
    """

    _taken_single_choices = []

    def __init__(self, _graph=None, _influence_matrix=None, _status_array=None, _choice_con_map=None,
                 _des_var_values=None, _metric_values=None, **_):
        self._graph = _graph or self._get_empty_graph()
        self._choice_constraints: List[ChoiceConstraint] = _choice_con_map or []
        self._influence_matrix: Optional[InfluenceMatrix] = _influence_matrix
        self._status_array: Optional[np.ndarray] = _status_array

        self._update_connector_grouping_degrees()
        self._des_var_values: Dict[DesignVariableNode, Union[float, int]] = (_des_var_values or {}).copy()
        self._metric_values: Dict[MetricNode, float] = (_metric_values or {}).copy()

    @staticmethod
    def _get_empty_graph():
        graph = nx.MultiDiGraph()
        # We need hashable edge attribute dicts to be able to use edges in sets
        graph.edge_attr_dict_factory = HashableDict
        return graph

    def __hash__(self):
        start_nodes = tuple(sorted(self.derivation_start_nodes, key=lambda n: n.name)) \
            if self.derivation_start_nodes else None

        g = self.graph
        node_hashes = tuple(sorted([hash(node) for node in g.nodes]))
        edge_hashes = tuple(sorted([hash(edge) for edge in g.edges]))

        constraints_hashes = tuple(hash(c) for c in self._choice_constraints)

        return hash((start_nodes, node_hashes, edge_hashes, constraints_hashes))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def fingerprint(self):
        """Like a hash but supports pickled nodes (note: not all node properties are compared!)"""

        def _node_fingerprint(node: DSGNode):
            return hash(node.str_context())

        def _edge_fingerprint(edge):
            return hash(tuple(_node_fingerprint(v) if i < 2 else v for i, v in enumerate(edge)))

        nodes_fingerprints = hash(tuple(sorted(_node_fingerprint(n) for n in self.graph.nodes)))
        edges_fingerprints = hash(tuple(sorted(_edge_fingerprint(e) for e in self.graph.edges)))

        start_fp = hash(tuple(sorted(_node_fingerprint(n) for n in (self.derivation_start_nodes or []))))
        constraint_fps = hash(tuple(hash((
            cc.type.name,
            tuple(_node_fingerprint(n) for n in cc.nodes),
            None if cc.options is None else tuple(_node_fingerprint(v) if isinstance(v, DSGNode) else v
                                                  for opt_list in cc.options for v in opt_list),
        )) for cc in self._choice_constraints))

        return hash((start_fp, nodes_fingerprints, edges_fingerprints, constraint_fps))

    def is_same(self, other: 'DSGType') -> bool:
        """Compare based on node export str and edge types (supports pickled nodes)"""

        if len(self.graph.nodes) != len(other.graph.nodes):
            return False
        if len(self.graph.edges) != len(other.graph.edges):
            return False

        return self.fingerprint() == other.fingerprint()

    """#####################################
    ### DERIVATION-START NODES FUNCTIONS ###
    #####################################"""

    @property
    def derivation_start_nodes(self) -> Optional[Set[DSGNode]]:
        """Set of nodes (if set) where node and choice derivation starts"""
        return self._get_derivation_start_nodes()

    @property
    def derivation_start_permanent_nodes(self) -> Optional[Set[DSGNode]]:
        derivation_start = self.derivation_start_nodes
        if derivation_start is None:
            return

        permanent_nodes = self._get_permanent_nodes()
        if permanent_nodes is not None:
            return derivation_start | permanent_nodes
        return derivation_start

    def _get_derivation_start_nodes(self) -> Optional[Set[DSGNode]]:
        raise NotImplementedError

    def _get_permanent_nodes(self) -> Optional[Set[DSGNode]]:
        pass

    """################################
    ### UNDERLYING GRAPHS FUNCTIONS ###
    ################################"""

    @property
    def graph(self):
        """The underlying networkx directed graph"""
        return self._graph

    def export_gml(self, path=None):
        """Export to GML (use e.g. Gephi to view)"""
        return export_gml(self._get_graph_for_export(), path)

    def export_dot(self, path=None, return_dot=False):
        """Export to DOT (use Graphviz / https://viz-js.com/ to view)"""
        return export_dot(self._get_graph_for_export(), path, start_nodes=self.derivation_start_nodes,
                          choice_constraints=self._choice_constraints, return_dot=return_dot)

    def export_drawio(self, path=None):
        """Export to draw.io"""
        return export_drawio(self._get_graph_for_export(), path, start_nodes=self.derivation_start_nodes,
                             choice_constraints=self._choice_constraints)

    def render(self, title=None, print_svg=False, print_dot=False):
        """Render the DSG as a graph.
        Open a webbrowser, or renders it as cell output if used in an IPython/Jupyter notebook."""
        from adsg_core.render import DSGRenderer
        DSGRenderer(self, title=title).render(print_svg=print_svg, print_dot=print_dot)

    def render_all(self, idx=None, title=None, print_svg=False, print_dot=False):
        """Renders all DSG instances as a graph (optionally provide only specific indices to render).
        Open a webbrowser, or renders it as cell output if used in an IPython/Jupyter notebook."""
        from adsg_core.render import DSGRenderer
        DSGRenderer(self, title=title).render_all_instances(idx=idx, print_svg=print_svg, print_dot=print_dot)

    def render_legend(self, elements=None):
        from adsg_core.render import DSGRenderer
        DSGRenderer(self).render_legend(elements=elements)

    def _get_graph_for_export(self):
        graph = nx.MultiDiGraph()
        graph.add_edges_from(self._graph.edges(data=True))
        graph.add_nodes_from(self._graph.nodes(data=True))

        for node in self.des_var_nodes:
            node.assigned_value = self.des_var_value(node)
        for node in self.metric_nodes:
            node.assigned_value = self.metric_value(node)

        self._export_prepare(graph)
        return graph

    def _export_prepare(self, graph: nx.MultiDiGraph):
        pass

    """#########################
    ### NODE TYPE PROPERTIES ###
    #########################"""

    @property
    def choice_nodes(self):
        return get_choice_nodes(self._graph)

    def get_nodes_by_type(self, type_):
        """Get all nodes in the graph that exactly match the given type (ignores subtypes!)"""
        return get_nodes_by_type(self._graph, type_)

    def get_nodes_by_subtype(self, type_):
        """Get all nodes in the graph that are or inherit from the given type"""
        return get_nodes_by_subtype(self._graph, type_)

    """##########################################
    ### DESIGN VARIABLE AND METRICS FUNCTIONS ###
    ##########################################"""

    @property
    def all_des_var_nodes(self) -> List[DesignVariableNode]:
        """Get all design variables nodes in the graph"""
        return self.ordered_choice_nodes(self.get_nodes_by_type(DesignVariableNode))

    @property
    def des_var_nodes(self) -> List[DesignVariableNode]:
        """Get all non-constrained (non-linked) design variable nodes in the graph"""
        seen_constraint_sets = set()
        mirror_con_dvs = {node: i for i, con in enumerate(self._choice_constraints) for node in con.nodes}

        # Only take the first-seen nodes out of each design constraint set
        des_var_nodes = []
        for node in self.all_des_var_nodes:
            i_set = mirror_con_dvs.get(node)
            if i_set is not None:
                if i_set in seen_constraint_sets:
                    continue
                seen_constraint_sets.add(i_set)

            des_var_nodes.append(node)
        return des_var_nodes

    def set_des_var_value(self, des_var_node: DesignVariableNode, value: Union[float, int]):
        """
        Set the value of a design variable node.
        Discrete values are defined as an option index (so not the value of an option itself).
        """
        value, bounds_fraction = des_var_node.correct_value(value)
        self._des_var_values[des_var_node] = value

        # Set values of linked design variables
        decision_constraint = self.is_constrained_choice(des_var_node)
        if decision_constraint is not None:
            if decision_constraint.type != ChoiceConstraintType.LINKED:
                raise RuntimeError(f'Unsupported constraint type for DV {des_var_node!r}: {decision_constraint.type}')
            for linked_des_var_node in decision_constraint.nodes:
                if linked_des_var_node == des_var_node:
                    continue

                # Check that design variables are of the same type
                if linked_des_var_node.is_discrete != des_var_node.is_discrete:
                    raise ValueError(f'Dependent des vars should be same type: discrete = {des_var_node.is_discrete}')

                # Map value of taken design variable onto domain of linked design variable
                if des_var_node.is_discrete:
                    dep_value = value
                else:
                    if linked_des_var_node.bounds is None:
                        raise ValueError(f'Design variable bounds not set: {linked_des_var_node!r}')
                    dep_lower, dep_upper = linked_des_var_node.bounds
                    dep_value = dep_lower + bounds_fraction * (dep_upper - dep_lower)

                self._des_var_values[linked_des_var_node] = dep_value

    def des_var_value(self, des_var_node: DesignVariableNode) -> Optional[Union[float, int]]:
        """
        Get the value of a design variable node. If not set, returns None.
        For discrete design variables, the index of the selected option is returned.
        """
        return self._des_var_values.get(des_var_node)

    @property
    def des_var_values(self) -> Dict[DesignVariableNode, Union[float, int]]:
        return self._des_var_values.copy()

    def reset_des_var_values(self):
        self._des_var_values = {}

    @property
    def metric_nodes(self) -> List[MetricNode]:
        return self.get_nodes_by_type(MetricNode)

    def set_metric_value(self, metric_node: MetricNode, value: float):
        """
        Set the value of a metric node.
        """
        self._metric_values[metric_node] = value

    def metric_value(self, metric_node) -> Optional[float]:
        return self._metric_values.get(metric_node)

    @property
    def metric_values(self):
        return self._metric_values.copy()

    def reset_metric_values(self):
        self._metric_values = {}

    """################################
    ### CHOICE CONSTRAINT FUNCTIONS ###
    ################################"""

    def is_constrained_choice(self, choice_node: CDVNode) -> Optional[ChoiceConstraint]:
        for choice_con in self._choice_constraints:
            if choice_node in choice_con.nodes:
                return choice_con

    def get_choice_constraints(self) -> List[ChoiceConstraint]:
        return self._choice_constraints

    def constrain_choices(self, constraint: ChoiceConstraintType, choice_nodes: List[CDVNode],
                          remove_infeasible_choices=True) -> 'DSG':
        """
        Create a new choice constraint (see `ChoiceConstraintType`).
        For design variable nodes, only the LINKED constraint is available.

        Following constraints are available (`ChoiceConstraintType`):

        - `LINKED`: To make all choices have the same option index             --> AA, BB, CC
        - `PERMUTATION`: To make the choices be permutations of option indices --> AB, BA, AC, CA, BC, CB
        - `UNORDERED`: To have all option index combinations without ordering  --> AA, AB, AC, BB, BC, CC
        - `UNORDERED_NOREPL`: Same but also without replacement                --> AB, AC, BC
        """
        if len(choice_nodes) == 0:
            return self
        choice_nodes = self.ordered_choice_nodes(choice_nodes)

        # Check if any nodes are previously constrained
        constrained_nodes = {node for choice_con in self._choice_constraints for node in choice_con.nodes}
        for node in choice_nodes:
            if node in constrained_nodes:
                raise RuntimeError(f'Node is already constrained: {node}')

        # Check constraint types
        node_type = type(choice_nodes[0])
        for node in choice_nodes:
            if type(node) != node_type:
                raise ValueError(f'Choice nodes should all be of type: {node_type}')
            if isinstance(node, ConnectionChoiceNode):
                raise RuntimeError(f'Constraints not implemented for connection choices!')
            if isinstance(node, DesignVariableNode) and constraint != ChoiceConstraintType.LINKED:
                raise ValueError(f'Only LINKED constraint can be applied to DV nodes: {node}')

        # Add the constraint
        node_options = []
        is_selection_choice = False
        for node in choice_nodes:
            if isinstance(node, SelectionChoiceNode):
                is_selection_choice = True
                node_options.append(self.get_option_nodes(node))
                continue

            # if isinstance(node, ConnectionChoiceNode):
            #     node_options.append(list(node.iter_conn_permutations(self)))
            #     continue

        if len(node_options) == 0:
            node_options = None
        elif constraint in [ChoiceConstraintType.UNORDERED, ChoiceConstraintType.UNORDERED_NOREPL]:
            n_opts = len(node_options[0])
            if any([len(opts) != n_opts for opts in node_options]):
                raise ValueError(f'All choices must have the same nr of options for constraint type: {constraint}')

        decision_constraint = ChoiceConstraint(constraint, choice_nodes, node_options)
        self._choice_constraints.append(decision_constraint)

        # Remove constrained selection-choices that would always lead to infeasible architectures
        if is_selection_choice:
            _, init_choice_nodes = traverse_until_choice_nodes(self.graph, self.derivation_start_permanent_nodes)
            permanent_nodes = {node for node in choice_nodes if node in init_choice_nodes}

            has_removed = False
            for dec_node, removed_opts in get_constraint_pre_removed_options(decision_constraint, permanent_nodes):
                removed_edges, removed_nodes = \
                    get_mod_apply_choice_constraint(self._graph, self.derivation_start_nodes, dec_node, removed_opts)
                self._graph.remove_edges_from(removed_edges)
                self._graph.remove_nodes_from(removed_nodes)

                if len(removed_opts) > 0:
                    has_removed = True
            if has_removed and remove_infeasible_choices:
                return self.resolve_single_selection_choices()

        return self

    def _get_removed_constrained_selection_choices(
            self, sel_choice_node: SelectionChoiceNode, option_node: Optional[DSGNode]) \
            -> List[Tuple[SelectionChoiceNode, List[DSGNode]]]:
        choice_constraint = self.is_constrained_choice(sel_choice_node)
        if choice_constraint is None or option_node is None:
            return []

        # Get index of decision and chosen option node
        for i_dec, dec_node in enumerate(choice_constraint.nodes):
            if dec_node == sel_choice_node:
                try:
                    i_opt = choice_constraint.options[i_dec].index(option_node)
                    break
                except ValueError:
                    raise ValueError(f'Node {option_node!r} not an option of choice: {sel_choice_node!r}')
        else:
            raise RuntimeError(f'Choice node not found in constraint: {sel_choice_node!r}')

        # Determine options of other choices to remove
        removed_sel_choice_opts: List[Tuple[SelectionChoiceNode, List[DSGNode]]]
        removed_sel_choice_opts = get_constraint_removed_options(choice_constraint, i_dec, i_opt)
        return removed_sel_choice_opts

    """###########################
    ### CHOICE NODES FUNCTIONS ###
    ###########################"""

    def get_ordered_next_choice_nodes(self) -> List[ChoiceNode]:
        """Get the next (selection or connection) choice node(s) that are active and can be applied"""
        if self._influence_matrix is None:
            self.set_influence_matrix()
        if self._influence_matrix is None or self._status_array is None:
            return []

        return self._influence_matrix.get_next_choice_nodes(self._status_array)

    def ordered_choice_nodes(self, choice_nodes):
        return sorted(choice_nodes, key=self._choice_sort_key)

    def _choice_sort_key(self, choice_node: CDVNode) -> tuple:
        raise NotImplementedError

    def get_confirmed_graph(self):
        """Get a subset of the DSG only containing confirmed nodes (i.e. everything up to the next active choices)"""
        start_nodes: Set[DSGNode] = self.derivation_start_permanent_nodes
        if start_nodes is None:
            start_nodes = self._get_alternative_start_nodes()
        non_confirmed_nodes = get_non_confirmed_nodes(self._graph, start_nodes)
        return self.get_for_adjusted(removed_nodes=non_confirmed_nodes)

    def _get_alternative_start_nodes(self) -> Set[DSGNode]:
        raise NotImplementedError

    def get_originating_node(self, choice_node: ChoiceNode) -> DSGNode:
        return list(self._graph.predecessors(choice_node))[0]

    def get_option_nodes(self, choice_node: SelectionChoiceNode):
        """Get a list of option nodes available for a selection choice"""
        return sorted(self._graph.successors(choice_node),
                      key=lambda n: (str(n.decision_id or ''), n.option_id))

    def iter_possible_connection_edges(self, choice_node: ConnectionChoiceNode):
        """Iterate over possible sets of connection edges for this connection choice"""
        yield from choice_node.iter_conn_edges(self)

    def n_options(self, choice_node):
        if isinstance(choice_node, ConnectionChoiceNode):
            return len(list(choice_node.iter_conn_edges(self)))
        return len(self.get_option_nodes(choice_node))

    def _update_connector_grouping_degrees(self):
        node: ConnectorDegreeGroupingNode
        for node in self.get_nodes_by_type(ConnectorDegreeGroupingNode):
            node.update_deg(self._graph)

    """#####################
    ### CHOICE FUNCTIONS ###
    #####################"""

    def initialize_choices(self):
        """
        Initializes choice nodes and makes the DSG ready for use in deriving architecture instances:
        - Removes nodes that are incompatible with any of the initially-confirmed nodes
        - Sets the influence matrix, needed for determining choice order
        - Resolves initially-active choices with 0 or 1 options
        """
        dsg = self
        if not dsg.derivation_start_nodes:
            raise RuntimeError('Start nodes not set!')

        # Resolve incompatibility constraints
        try:
            removed_nodes = get_mod_nodes_remove_incompatibilities(dsg.graph, dsg.derivation_start_nodes)
            if len(removed_nodes) > 0:
                dsg = dsg.get_for_adjusted(removed_nodes=removed_nodes)

        # In case of problems, just return the new graph: it will be infeasible
        except IncompatibilityError:
            pass

        dsg.set_influence_matrix()
        return dsg.resolve_single_selection_choices()

    def resolve_single_selection_choices(self) -> 'DSG':
        """Removes selection-choices that have one or no options left"""
        graph = self
        taken_choices = []
        while True:
            for sel_choice_node in list(graph.get_ordered_next_choice_nodes()):
                if isinstance(sel_choice_node, SelectionChoiceNode) and sel_choice_node in graph.graph.nodes:
                    opt_nodes = graph.get_option_nodes(sel_choice_node)
                    if len(opt_nodes) <= 1:
                        option_node = opt_nodes[0] if len(opt_nodes) == 1 else None
                        taken_choices.append((sel_choice_node, option_node))
                        self.__class__._taken_single_choices = []

                        graph = graph.get_for_apply_selection_choice(sel_choice_node, option_node)
                        taken_choices += self.__class__._taken_single_choices
                        break
            else:
                break
        self.__class__._taken_single_choices = taken_choices
        return graph

    def set_influence_matrix(self):
        try:
            self._influence_matrix = InfluenceMatrix(self)
        except ValueError:
            pass
        if self._influence_matrix is not None:
            self._status_array = self._influence_matrix.init_status_array

    """###########################
    ### APPLY CHOICE FUNCTIONS ###
    ###########################"""

    def get_for_apply_selection_choice(self, choice_node, target_option_node):
        """Get a new DSG where the selection choice has been applied with the provided target option node"""
        if self.derivation_start_nodes is None:
            raise RuntimeError('Start nodes not set!')

        removed_edges, removed_nodes, added_edges = \
            self.get_mod_apply_selection_choice(choice_node, target_option_node)

        if self._status_array is None or self._influence_matrix is None:
            raise RuntimeError('Influence matrix not set: check graph is feasible and has decisions!')
        status_array = self._influence_matrix.apply_selection_choice(
            self._status_array, choice_node, target_option_node)

        graph = self.get_for_adjusted(removed_edges=removed_edges, removed_nodes=removed_nodes,
                                      added_edges=added_edges, status_array=status_array)
        return graph.resolve_single_selection_choices()

    def get_taken_single_selection_choices(self) -> List[Tuple[SelectionChoiceNode, Optional[DSGNode]]]:
        return self.__class__._taken_single_choices

    def get_mod_apply_selection_choice(self, choice_node: SelectionChoiceNode, target_option_node,
                                       only_added=False) -> tuple:
        choice_constraint_map = self._get_removed_constrained_selection_choices(choice_node, target_option_node)
        return get_mod_apply_selection_choice(
            self._graph, self.derivation_start_nodes, choice_node, target_option_node, choice_constraint_map,
            only_added=only_added)

    def get_confirmed_edges_selection_choice(self, choice_node, target_option_node, include_choice=False,
                                             included_apply_edges=True, cache=None) \
            -> Tuple[Set[EdgeTuple], Set[DSGNode]]:

        _, _, added_edges = self.get_mod_apply_selection_choice(choice_node, target_option_node, only_added=True)
        confirmed_edges = get_confirmed_edges_for_node(
            self._graph, target_option_node, include_choice=include_choice, cache=cache)
        if included_apply_edges:
            confirmed_edges |= added_edges

        # Get nodes removed by incompatibility constraints
        confirmed_start_nodes = self.derivation_start_nodes | {target_option_node}
        removed_nodes = get_mod_nodes_remove_incompatibilities(self._graph, confirmed_start_nodes, cache=cache)

        return confirmed_edges, removed_nodes

    def get_for_apply_connection_choices(
            self, choice_node_edges: List[Tuple[ConnectionChoiceNode, Sequence[Tuple[ConnectorNode, ConnectorNode]]]],
            validate=True) -> 'DSG':
        """Get a new DSG where the connection choices have the provided connection edges applied"""
        if self.derivation_start_nodes is None:
            raise RuntimeError('Start nodes not set!')

        removed_edges = set()
        removed_nodes = set()
        added_edges = set()
        status_array = self._status_array.copy()

        for choice_node, edges in choice_node_edges:
            if validate and len(edges) > 0 and not choice_node.validate_conn_edges(self, edges):
                raise ValueError(f'Invalid edges for connection choice ({choice_node!r}): {edges!r}')

            removed_edges_i, removed_nodes_i, added_edges_i = self.get_mod_apply_connection_choice(choice_node, edges)
            removed_edges = removed_edges.union(removed_edges_i)
            removed_nodes = removed_nodes.union(removed_nodes_i)
            added_edges = added_edges.union(added_edges_i)

            status_array = self._influence_matrix.apply_choice(status_array, choice_node, copy=False)

        return self.get_for_adjusted(removed_edges=removed_edges, removed_nodes=removed_nodes,
                                     added_edges=added_edges, status_array=status_array)

    def get_for_apply_connection_choice(self, choice_node: ConnectionChoiceNode,
                                        edges: Sequence[Tuple[ConnectorNode, ConnectorNode]] = None, validate=True):
        """Get a new DSG where the connection choice has the provided connection edges applied"""
        if edges is None:
            edges = tuple()
        return self.get_for_apply_connection_choices([(choice_node, edges)], validate=validate)

    def get_mod_apply_connection_choice(self, choice_node: ConnectionChoiceNode,
                                        edges: Sequence[Tuple[ConnectorNode, ConnectorNode]]) -> tuple:
        return get_mod_apply_connection_choice(self._graph, choice_node, edges)

    def get_confirmed_edges_connection_choice(self, choice_node, connection_option):
        _, _, added_edges = self.get_mod_apply_connection_choice(choice_node, connection_option)
        return added_edges

    """###########################
    ### GRAPH STATUS FUNCTIONS ###
    ###########################"""

    @property
    def feasible(self):
        """Whether the architecture is feasible or not (unconnectable connectors and/or unsolvable incompatibilities)"""
        try:
            self._check_unconnected_connectors()
        except ValueError:
            return False
        if self.has_confirmed_incompatibility_edges():
            return False
        return True

    @property
    def final(self):
        """Whether the architecture is final (no more choices to make) or not"""
        return len(self._graph.nodes) > 0 and len(self.choice_nodes) == 0

    def _check_unconnected_connectors(self):
        if len(get_unconnected_connectors(self._graph, self.derivation_start_nodes, stop_at_one=True)) > 0:
            raise ValueError('There are unconnectable ports')

    @property
    def unconnected_connectors(self):
        return get_unconnected_connectors(self._graph, self.derivation_start_nodes)

    """#############################
    ### GRAPH QUERYING FUNCTIONS ###
    #############################"""

    def has_conditional_existence(self, target_node, base_nodes=None):
        return has_conditional_existence(self._graph, self.derivation_start_nodes, target_node, base_nodes=base_nodes)

    def derives(self, source_node: DSGNode, target_node: DSGNode, connects=False):
        """
        Check whether the source node derives the target node (i.e. whether there is a path along derivation edges from
        source to target). If `connects` is set to True, also connection edges can be part of a path.
        """
        return check_derives(self._graph, source_node, target_node, connects=connects)

    """############################
    ### GRAPH COPYING FUNCTIONS ###
    ############################"""

    def copy(self):
        """Create a copy of the DSG"""
        return self.get_for_adjusted()

    def get_for_adjusted(self, removed_edges=None, removed_nodes=None, added_edges=None,
                         status_array=None, inplace=False, **kwargs):
        graph = self._graph
        graph_copy = graph if inplace else self._get_empty_graph()

        if not inplace:
            graph_copy.add_edges_from(graph.edges(keys=True, data=True))
            graph_copy.add_nodes_from(graph.nodes)
        if removed_edges is not None:
            graph_copy.remove_edges_from(removed_edges)
        if removed_nodes is not None:
            graph_copy.remove_nodes_from(removed_nodes)
        if added_edges is not None:
            graph_copy.add_edges_from(added_edges)

        self._mod_graph_adjust_kwargs(kwargs)
        if inplace:
            if status_array is not None:
                self._status_array = status_array
            self._mod_graph_inplace(kwargs)
            return self

        dec_con_map_copy = self._choice_constraints.copy()
        return self.__class__(_graph=graph_copy, _influence_matrix=self._influence_matrix,
                              _status_array=status_array if status_array is not None else self._status_array,
                              _choice_con_map=dec_con_map_copy, _des_var_values=self._des_var_values,
                              _metric_values=self._metric_values, **kwargs)

    def _mod_graph_adjust_kwargs(self, kwargs):
        pass

    def _mod_graph_inplace(self, kwargs):
        pass

    def get_for_kept_edges(self, kept_edges, **kwargs):
        graph_copy = self._get_empty_graph()
        graph_copy.add_edges_from(kept_edges)

        self._mod_graph_adjust_kwargs(kwargs)
        return self.__class__(_graph=graph_copy, _influence_matrix=self._influence_matrix,
                              _status_array=self._status_array, _choice_con_map=self._choice_constraints,
                              _des_var_values=self._des_var_values, _metric_values=self._metric_values, **kwargs)

    """#########################################
    ### INCOMPATIBILITY CONSTRAINT FUNCTIONS ###
    #########################################"""

    def add_incompatibility_constraint(self, nodes: List[DSGNode], **attr):
        """
        An incompatibility constraint defines that two nodes cannot exist in an architecture at the same time. This
        means that when one of the nodes becomes confirmed by making a choice, the other node (and its derived nodes)
        will be removed from the graph. If both of the nodes are confirmed, it means that there is no way to have only
        one of them in an architecture, and that therefore the architecture is infeasible (this can also happen before
        any choices have been made!).

        Incompatibility constraints should be added before initializing the graph's choices.
        """
        add_incompatibility_constraint(self._graph, nodes, **attr)
        return self

    def get_confirmed_incompatibility_edges(self) -> Set[EdgeTuple]:
        """Returns incompatibility edges that link two confirmed nodes. There should be none in order to have a
        feasible architecture."""
        if self.derivation_start_nodes is None:
            return set()
        return get_confirmed_incompatibility_edges(self._graph, self.derivation_start_nodes)

    def has_confirmed_incompatibility_edges(self):
        return len(self.get_confirmed_incompatibility_edges()) > 0

    def get_incompatibility_deriving_nodes(self, target_node: DSGNode, confirmed_nodes: Set[DSGNode]):
        return get_incompatibility_deriving_nodes(self._graph, target_node, confirmed_nodes)


DSGType = TypeVar('DSGType', bound=DSG)

ADSG = DSG  # Backward compatibility
ADSGType = TypeVar('ADSGType', bound=ADSG)  # Backward compatibility



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import warnings
import itertools
import numpy as np
from typing import *
import networkx as nx
from io import BytesIO, StringIO
from collections import defaultdict
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *
from adsg_core.graph.choice_constraints import *

__all__ = ['export_gml', 'export_dot', 'export_drawio']


def export_gml(graph: nx.MultiDiGraph, path: str = None):
    fp = BytesIO() if path is None else path
    nx.write_gml(graph, fp, stringizer=str)
    return fp.getvalue().decode('utf-8') if path is None else None


def export_dot(graph: nx.MultiDiGraph, path=None, start_nodes: Set[DSGNode] = None,
               choice_constraints: List[ChoiceConstraint] = None, legend_mode=False, return_dot=False):
    graph_export = nx.DiGraph()

    if start_nodes is None:
        start_nodes = set()

    shape_map = {
        NodeExportShape.CIRCLE: 'ellipse',
        NodeExportShape.ROUNDED_RECT: 'rect',
        NodeExportShape.HEXAGON: 'hexagon',
    }
    node_map = {}

    def get_node(node: DSGNode, node_id):
        dot_node_id = f'leg_{node_id}' if legend_mode else node_id
        if node_id not in node_map:
            label = str(node.get_export_title())
            style = ['filled']
            if node in start_nodes:
                style.append('bold')
                label = f'<<B>{label}</B>>'
            else:
                label = '"'+label+'"'

            color = node.get_export_color()
            graph_export.add_node(
                dot_node_id, label=label,
                style='"'+','.join(style)+'"', fillcolor=color,
                shape=shape_map.get(node.get_export_shape(), 'ellipse'),
                margin=0 if legend_mode else 0.05,
            )

        return dot_node_id

    i = 0
    node_id_map = {}

    # Hack to place the choice constraints lower in the graph
    if legend_mode:
        for choice_constraint in (choice_constraints or []):
            for node in choice_constraint.nodes:
                if node not in node_id_map:
                    node_id_map[node] = i
                    i += 1
                get_node(node, node_id_map[node])

    shown_incompatibilities = set()
    repeated_conn_edges = defaultdict(int)
    u: DSGNode
    v: DSGNode
    for u, v, k, d in graph.edges(keys=True, data=True):
        if u not in node_id_map:
            node_id_map[u] = i
            i += 1
        if v not in node_id_map:
            node_id_map[v] = i
            i += 1

        u_node = get_node(u, node_id_map[u])
        v_node = get_node(v, node_id_map[v])

        attr = {}
        for key, value in d.items():
            value = str(value)
            if ':' in value:
                value = f'\"{value}\"'
            attr[key] = value

        edge_str = None
        edge_type = get_edge_type((u, v, k, d))
        if edge_type == EdgeType.INCOMPATIBILITY:
            attr['color'] = 'red'
            attr['arrowhead'] = 'none'
            if not legend_mode:
                attr['constraint'] = 'false'

            if (u, v) in shown_incompatibilities:
                continue
            shown_incompatibilities |= {(u, v), (v, u)}

        elif edge_type == EdgeType.CONNECTS:
            attr['style'] = 'dashed'
        elif edge_type == EdgeType.EXCLUDES:
            attr['style'] = 'dashed'
            attr['color'] = 'red'
            if legend_mode:
                edge_str = 'connection\\nexcluded'

        if isinstance(u, ConnectorNode):
            if ((edge_type == EdgeType.DERIVES and isinstance(v, ConnectorDegreeGroupingNode)) or
                    (edge_type == EdgeType.CONNECTS and isinstance(v, ConnectionChoiceNode))):
                edge_str = u.get_full_deg_str()
        elif isinstance(u, ConnectionChoiceNode) and edge_type == EdgeType.CONNECTS and isinstance(v, ConnectorNode):
            edge_str = v.get_full_deg_str()
        if edge_str is not None:
            attr['label'] = '"'+edge_str+'"'

        if edge_type == EdgeType.CONNECTS:
            repeated_conn_edges[(u_node, v_node)] += 1

        graph_export.add_edge(u_node, v_node, **attr)

    # Set text of repeated connection edges
    for (u_node, v_node), count in repeated_conn_edges.items():
        if count > 1:
            nx.set_edge_attributes(graph_export, {(u_node, v_node): {'label': f'"{count}x"'}})

    # Add individual nodes
    for node_ in graph.nodes:
        if node_ not in node_id_map:
            node_id_map[node_] = i
            i += 1

            get_node(node_, node_id_map[node_])

    # Add choice constraints
    for choice_constraint in (choice_constraints or []):
        for u, v in itertools.combinations(choice_constraint.nodes, 2):
            if u not in node_id_map or v not in node_id_map:
                continue
            u_node = get_node(u, node_id_map[u])
            v_node = get_node(v, node_id_map[v])

            attr = {
                'color': '#9C27B0',
                'style': 'dotted',
                'penwidth': '3',
                'arrowhead': 'none',
                'constraint': 'false',
                'label': CCT_EXPORT_LABEL.get(choice_constraint.type, choice_constraint.type.name),
            }
            if legend_mode:
                del attr['constraint']
                attr['label'] = 'choice\\nconstraint'
            graph_export.add_edge(u_node, v_node, **attr)

    warnings.filterwarnings('ignore', message=r'.*write\_dot.*', category=PendingDeprecationWarning)

    # Define graph attributes (weird notation but works)
    graph_export.graph['graph'] = dict(
        rankdir='LR',  # Arrange left-to-right (vs vertical)
        dpi='60',
        fontsize='20',
        **(dict(
            ranksep=0,
            nodesep=.1,
        ) if legend_mode else {}),
    )
    if legend_mode:
        graph_export.graph['node'] = dict(
            height=0,
        )

    dot_graph = nx.nx_pydot.to_pydot(graph_export)
    if return_dot:
        return dot_graph

    dot_export = dot_graph.to_string()
    if path is not None:
        with open(path, 'w') as fp:
            fp.write(dot_export)
    return dot_export


def export_drawio(graph: nx.MultiDiGraph, path: str = None, start_nodes: Set[DSGNode] = None,
                  choice_constraints: List[ChoiceConstraint] = None):
    from lxml.builder import E
    import lxml.etree as etree

    if start_nodes is None:
        start_nodes = set()

    shape_map = {
        NodeExportShape.CIRCLE: 'ellipse',
        NodeExportShape.ROUNDED_RECT: 'rounded=1',
        NodeExportShape.HEXAGON: 'shape=hexagon;perimeter=hexagonPerimeter2;fixedSize=1;size=10',
    }

    cells = [
        E.mxCell(id="0"), E.mxCell(id="1", parent="0"),
    ]

    cell_id = 2
    cell_id_map = {}
    height = 20
    geom = {'as': 'geometry'}
    for node in graph.nodes:
        if not isinstance(node, DSGNode):
            continue

        style = [
            shape_map[node.get_export_shape()], f'fillColor={node.get_export_color()}',
            'whiteSpace=wrap', 'html=1',
        ]
        if node in start_nodes:
            style.append('strokeWidth=3')

        x = np.round(np.random.random()*50)*10
        y = np.round(np.random.random()*50)*10
        width = height
        if node.get_export_shape() == NodeExportShape.HEXAGON:
            width = 1.5*height

        cells.append(E.mxCell(
            E.mxGeometry(x=str(x), y=str(y), width=str(width), height=str(height), **geom),
            id=str(cell_id), value=node.get_export_title(), style=';'.join(style), vertex='1', parent='1',
        ))
        cell_id_map[node] = str(cell_id)
        cell_id += 1

    def _add_edge(src_id_, tgt_id_, style_, edge_str_):
        nonlocal cell_id
        cells.append(E.mxCell(
            E.mxGeometry(relative='1', **geom),
            id=str(cell_id), style=';'.join(style_), edge='1', parent='1', source=src_id_, target=tgt_id_,
        ))
        edge_id = cell_id
        cell_id += 1

        if edge_str_ is not None:
            style__ = 'edgeLabel;html=1;align=center;verticalAlign=middle;resizable=0'
            cells.append(E.mxCell(
                E.mxGeometry(relative='1', **geom),
                id=str(cell_id), value=edge_str_, style=style__, parent=str(edge_id), vertex='1', connectable='0',
            ))
            cell_id += 1

    default_edge_style = [
        'rounded=1', 'orthogonalLoop=1', 'jettySize=auto', 'html=1',
    ]
    for edge in iter_edges(graph):
        src, tgt = edge[0], edge[1]
        if src not in cell_id_map or tgt not in cell_id_map:
            continue
        src_id, tgt_id = cell_id_map[src], cell_id_map[tgt]
        edge_type = get_edge_type(edge)

        style = default_edge_style.copy()
        if edge_type in [EdgeType.CONNECTS, EdgeType.EXCLUDES]:
            style.append('dashed=1')
        if edge_type in [EdgeType.INCOMPATIBILITY, EdgeType.EXCLUDES]:
            style.append('strokeColor=#ff0000')
        if edge_type == EdgeType.INCOMPATIBILITY:
            style.append('endArrow=none;endFill=0')

        edge_str = None
        if isinstance(src, ConnectorNode):
            if (isinstance(tgt, ConnectorDegreeGroupingNode) or
                    (edge_type == EdgeType.CONNECTS and isinstance(tgt, ConnectionChoiceNode))):
                edge_str = src.get_full_deg_str()
        elif (isinstance(src, ConnectionChoiceNode) and edge_type == EdgeType.CONNECTS and
              isinstance(tgt, ConnectorNode)):
            edge_str = tgt.get_full_deg_str()

        _add_edge(src_id, tgt_id, style, edge_str)

    # Add choice constraints
    for choice_constraint in (choice_constraints or []):
        for src, tgt in itertools.combinations(choice_constraint.nodes, 2):
            if src not in cell_id_map or tgt not in cell_id_map:
                continue
            src_id, tgt_id = cell_id_map[src], cell_id_map[tgt]

            style = default_edge_style.copy()
            style += [
                'dashed=1',
                'strokeColor=#880E4F',
                'endArrow=none;endFill=0',
            ]
            label = CCT_EXPORT_LABEL.get(choice_constraint.type, choice_constraint.type.name)
            _add_edge(src_id, tgt_id, style, label)

    root = etree.ElementTree(E.mxGraphModel(
        E.root(*cells),
    ))

    xml_contents = etree.tostring(root, encoding='utf-8', pretty_print=True)
    if path is None:
        return xml_contents
    with open(path, 'wb') as fp:
        fp.write(xml_contents)



"""
MIT License

Copyright: (c) 2023, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
from typing import *
from cached_property import cached_property
from adsg_core.optimization.sel_choice_enc.encoder import *

__all__ = ['BruteForceSelectionChoiceEncoder']


class BruteForceSelectionChoiceEncoder(SelectionChoiceEncoder):
    """
    Reference implementation of the selection-choice encoder: derives all relevant information from
    exhaustively-generated design vectors and node statuses. Works correctly for small problems but quickly runs into
    memory and time constraints for larger problems.
    """

    def _get_design_variable_options(self) -> List[int]:
        """Get for each design variable the number of discrete options (at least 2)"""
        x_all, _ = self.all_design_vectors_and_statuses
        if x_all.shape[0] == 0 or x_all.shape[1] == 0:
            return []
        return list(np.max(x_all, axis=0)+1)

    def _design_variable_is_forced(self) -> List[bool]:
        """Get for each design variable whether it is forced or not"""
        if self.n_valid == 0:
            return []

        # Loop over each choice
        x_all, _ = self.all_design_vectors_and_statuses
        x_all_imputed = x_all.copy()
        x_all_imputed[x_all_imputed == X_INACTIVE_VALUE] = 0
        is_forced = [False for _ in range(x_all.shape[1])]
        for i_x in range(1, x_all.shape[1]):
            # Get unique values
            values, idx = np.unique(x_all_imputed[:, i_x], return_inverse=True)

            # Check if any of the left-side design variable values for each of the unique values overlaps
            all_left_side_dvs = set()
            for i_value in range(len(values)):
                dv_idx = np.where(idx == i_value)[0]
                left_side_dvs = {hash(tuple(dv)) for dv in x_all_imputed[dv_idx, :i_x]}
                if len(left_side_dvs & all_left_side_dvs) > 0:
                    break
                all_left_side_dvs |= left_side_dvs
            else:
                is_forced[i_x] = True

        return is_forced

    def _count_n_valid(self) -> int:
        """Get the number of valid design vectors"""
        return self.all_design_vectors_and_statuses[0].shape[0]

    def get_design_vector_indices(self, choice_opt_idx_map: Dict[int, int]) -> Set[int]:
        """
        Get design vectors indices conforming to a subset of choice_idx-opt_idx selections.
        Note: inactive design variables are matched by any opt_idx
        """
        x_all, _ = self.all_design_vectors_and_statuses
        return self._match_dv_indices(x_all, choice_opt_idx_map)

    @staticmethod
    def _match_dv_indices(x_all: np.ndarray, choice_opt_idx_map: Dict[int, int]) -> Set[int]:
        if x_all.shape[1] == 0:
            return {0} if x_all.shape[0] == 1 else set()

        matched_dv_idx = None
        for i_choice, i_opt in choice_opt_idx_map.items():
            # Match exact values and inactive values
            i_matches = np.where((x_all[:, i_choice] == i_opt) | (x_all[:, i_choice] == X_INACTIVE_VALUE))[0]

            if matched_dv_idx is None:
                matched_dv_idx = set(i_matches)
            else:
                matched_dv_idx &= set(i_matches)

        return matched_dv_idx if matched_dv_idx is not None else set()

    def get_design_vector_index(self, opt_idx: List[int], i_dv_excluded: Set[int] = None) -> Tuple[List[int], int]:
        """
        Get the design vector and associated index that most closely represents the given selection of choice-opt_idx
        values. Optionally a set of not-allowed design vector indices can be given.
        If none is found, a RuntimeError should be thrown.
        """

        # Check inputs
        if len(opt_idx) != self.n_choices:
            raise ValueError(f'Incorrect nr of options, expected {self.n_choices}, got {len(opt_idx)}')
        x_all, _ = self.all_design_vectors_and_statuses

        # If there are no choices, return the first possible design vector
        if len(opt_idx) == 0:
            return list(x_all[0, :]), 0

        # Initialize the exclusion set
        if i_dv_excluded is None:
            i_dv_excluded = set()
        if len(i_dv_excluded) == self.n_valid:
            raise RuntimeError('All design vectors are excluded!')

        # Check if the design vector has an exact match, ignoring forced choices
        opt_idx = np.array(opt_idx, dtype=int)
        opt_idx_not_forced = opt_idx[self._is_non_forced]
        i_dv = self._dv_map.get(tuple(opt_idx_not_forced))
        if i_dv is not None and i_dv not in i_dv_excluded:
            return list(x_all[i_dv, :]), i_dv

        # Otherwise, find the closest design vector
        x_all_non_forced: np.ndarray = x_all[:, self._is_non_forced]
        is_inactive: np.ndarray = x_all_non_forced == X_INACTIVE_VALUE

        idx_match: np.ndarray = x_all_non_forced == opt_idx_not_forced
        idx_match[is_inactive] = True
        nr_match = np.sum(idx_match, axis=1)
        for i_excl in i_dv_excluded:
            nr_match[i_excl] = -1

        highest_match_mask = nr_match == np.max(nr_match)
        i_matched = np.where(highest_match_mask)[0]
        if len(i_matched) == 1:
            i_dv = i_matched[0]
            return list(x_all[i_dv, :]), i_dv

        # If multiple design vectors have the same nr of matched design variables, find the closest
        diff: np.ndarray = x_all_non_forced[highest_match_mask, :]-opt_idx_not_forced
        diff = diff.astype(float)
        diff[is_inactive[highest_match_mask, :]] = .25
        dist = np.sqrt(np.nansum(diff**2, axis=1))
        i_min_dist = i_matched[np.argmin(dist)]
        return list(x_all[i_min_dist, :]), i_min_dist

    @cached_property
    def _is_non_forced(self):
        return ~np.array(self.design_variable_is_forced, dtype=bool)

    @cached_property
    def _dv_map(self):
        x_all, _ = self.all_design_vectors_and_statuses
        non_forced = self._is_non_forced
        return {tuple(xi): i for i, xi in enumerate(x_all[:, non_forced])}

    def get_node_existence_mask(self, node_ids: List[int], i_dv: int = None) -> np.ndarray:
        """
        Returns whether for each design vector the given node_ids exist or not.
        Optionally returns the status only for one specific design vector index.
        """
        node_idx_map = self.matrix_nodes_map
        i_nodes = [node_idx_map[node_id] for node_id in node_ids]

        _, all_nodes_status = self.all_design_vectors_and_statuses
        nodes_status: np.ndarray = all_nodes_status[:, i_nodes]
        if i_dv is not None:
            nodes_status = nodes_status[i_dv, :]

        # The nodes exist when their status is confirmed
        return nodes_status == Diag.CONFIRMED.value

    def generate_all_design_vectors_and_statuses(self) -> Tuple[np.ndarray, np.ndarray]:
        """Generate all valid design vectors and status arrays"""
        node_idx = self.matrix_nodes_map
        opt_node_map = self.choice_opt_map

        base_matrix = self.influence_matrix

        opt_invalid_status = {Diag.REMOVED.value, Diag.INFEASIBLE_OPTION.value}
        opt_idx_map = {node_idx[choice_node_id]: [node_idx[opt_node_id] for opt_node_id in opt_node_ids]
                       for choice_node_id, opt_node_ids in opt_node_map.items()}

        choice_constraint_map = {(node_idx[choice_id], node_idx[opt_id]):
                                 [(node_idx[c_id], node_idx[o_id]) for c_id, o_id in removed_map]
                                 for (choice_id, opt_id), removed_map in self.choice_constraint_map.items()}

        initial_status = Diag.INITIAL.value
        confirmed_status = Diag.CONFIRMED.value
        removed_status = Diag.REMOVED.value
        choice_made_status = Diag.CHOICE_MADE.value

        # Get choice node dependency
        choice_node_ids = self.choice_node_ids
        n_choice_nodes = len(choice_node_ids)
        i_choice_nodes = sorted([node_idx[choice_node_id] for choice_node_id in choice_node_ids])

        # Get choice-option influence map
        confirmation_influence_map = {}
        removal_influence_map = {}
        for opt_node_ids in opt_node_map.values():
            for opt_node_id in opt_node_ids:
                i_opt_ = node_idx[opt_node_id]
                confirmation_influence_map[i_opt_] = base_matrix[i_opt_, :] == OffDiag.CONFIRMATION.value
                removal_influence_map[i_opt_] = base_matrix[i_opt_, :] == OffDiag.REMOVAL.value

        def _take_next(des_vector: np.ndarray, status_array: np.ndarray, removed_choice_options: set) \
                -> Tuple[List[np.ndarray], List[np.ndarray]]:
            # Get next choice to make: starting from left, the first choice with confirmed status
            choice_make_idx = np.where(status_array[i_choice_nodes] == confirmed_status)[0]

            # Check if there are any choices left
            if len(choice_make_idx) == 0:
                # Return final status
                return [des_vector.copy()], [status_array.copy()]

            choice_node_idx = choice_make_idx[0]
            i_choice = i_choice_nodes[choice_node_idx]

            # Get option nodes
            i_opt_nodes = _get_opt_nodes(status_array, removed_choice_options, i_choice)
            if len(i_opt_nodes) == 0:
                # If there are no option nodes, this architecture is infeasible
                return [], []

            # Branch into the different options
            branched_des_vector_arrays = []
            branched_status_arrays = []
            for opt_idx, i_opt in i_opt_nodes:
                # Apply selected option
                branched_opt_idx, branched_status, branched_removed_choice_options = \
                    _apply_option(des_vector, status_array, removed_choice_options,
                                  i_choice, i_opt, opt_idx, choice_node_idx)
                if branched_status is None:
                    continue

                # Branch into subsequent choice-options
                next_opt_idx_arrays, next_status_arrays = \
                    _take_next(branched_opt_idx, branched_status, branched_removed_choice_options)
                branched_des_vector_arrays += next_opt_idx_arrays
                branched_status_arrays += next_status_arrays

            return branched_des_vector_arrays, branched_status_arrays

        def _get_opt_nodes(status_array: np.ndarray, removed_choice_options: set, i_node) -> List[Tuple[int, int]]:
            return [(opt_idx, i_opt) for opt_idx, i_opt in enumerate(opt_idx_map[i_node])
                    if status_array[i_opt] not in opt_invalid_status and (i_node, i_opt) not in removed_choice_options]

        def _apply_option(des_vector: np.ndarray, status_array: np.ndarray, removed_choice_options: set,
                          i_choice: int, i_opt: int, opt_idx: int, choice_node_idx: int) \
                -> Tuple[Optional[np.ndarray], Optional[np.ndarray], Optional[set]]:
            des_vector: np.ndarray = des_vector.copy()
            status_array: np.ndarray = status_array.copy()
            removed_choice_options = set(removed_choice_options)

            # Apply the choice
            _do_apply_choice(status_array, i_choice, i_opt)
            _do_apply_choice_constraints(removed_choice_options, i_choice, i_opt)
            des_vector[choice_node_idx] = opt_idx

            # Apply any remaining choices with only one option
            for other_choice_idx in range(n_choice_nodes):
                if other_choice_idx == choice_node_idx:
                    continue

                # Check if this decision is active
                i_choice_other = i_choice_nodes[other_choice_idx]
                if status_array[i_choice_other] != confirmed_status:
                    continue

                # Check if it is a decision with zero or one options left
                i_sel_choice = _get_opt_nodes(status_array, removed_choice_options, i_choice_other)
                if len(i_sel_choice) <= 1:
                    if len(i_sel_choice) == 0:
                        # Infeasible architecture: node without options detected
                        return None, None, None

                    opt_idx_apply, i_opt = i_sel_choice[0]
                    _do_apply_choice(status_array, i_choice_other, i_opt)
                    _do_apply_choice_constraints(removed_choice_options, i_choice_other, i_opt)
                    des_vector[other_choice_idx] = opt_idx_apply

            return des_vector, status_array, removed_choice_options

        def _do_apply_choice(status_array: np.ndarray, i_choice_apply, i_opt_apply):
            # Apply confirmation and removal influences
            initial_mask = status_array == initial_status
            for influence_map, diag_value in [(confirmation_influence_map, confirmed_status),
                                              (removal_influence_map, removed_status)]:
                influence_mask = influence_map[i_opt_apply] & initial_mask
                status_array[influence_mask] = diag_value

            # Mark choice as made and option node as confirmed
            status_array[i_choice_apply] = choice_made_status
            status_array[i_opt_apply] = confirmed_status

        def _do_apply_choice_constraints(removed_choice_options: set, i_choice_apply, i_opt_apply):
            for removed_key in choice_constraint_map.get((i_choice_apply, i_opt_apply), []):
                removed_choice_options.add(removed_key)

        # Generate matrix alternatives
        base_opt_idx_array = -np.ones((n_choice_nodes,), dtype=int)
        base_status_array = np.diag(base_matrix)
        des_vector_list, status_array_list = _take_next(base_opt_idx_array, base_status_array, set())

        if len(status_array_list) == 0:
            des_vector_matrix = np.zeros((0, n_choice_nodes), dtype=int)
            status_matrix = np.zeros((0, base_matrix.shape[0]), dtype=int)
        else:
            des_vector_matrix = np.array(des_vector_list, dtype=int)
            status_matrix = np.array(status_array_list, dtype=int)
        return des_vector_matrix, status_matrix



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import enum
from typing import *
import networkx as nx
if False:
    from adsg_core.graph.adsg_nodes import *

__all__ = ['HashableDict', 'add_edge', 'get_edge', 'get_edge_for_type', 'get_edge_type', 'EdgeType',
           'iter_edges', 'iter_in_edges', 'iter_out_edges', 'get_edge_data']


class EdgeType(enum.Enum):
    DERIVES = 1
    CONNECTS = 2
    EXCLUDES = 3
    INCOMPATIBILITY = 4


class HashableDict(dict):
    """
    Hashable dict, from: https://stackoverflow.com/a/1151686
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._key_cache = None

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._key_cache = None

    def __delitem__(self, key):
        super().__delitem__(key)
        self._key_cache = None

    def __key(self):
        if self._key_cache is None:
            self._key_cache = hash(tuple((k, self[k]) for k in sorted(self)))
        return self._key_cache

    def __hash__(self):
        if self._key_cache is not None:
            return self._key_cache
        return self.__key()

    def __eq__(self, other):
        return self.__key() == other.__key()


def add_edge(graph: nx.MultiDiGraph, from_node, to_node, key=None, is_conn=False, edge_type=None, **attr):
    if edge_type is not None:
        data = get_edge_for_type(from_node, to_node, edge_type, key=key, **attr)[-1]
    else:
        data = get_edge(from_node, to_node, key=key, is_conn=is_conn, **attr)[-1]
    graph.add_edge(from_node, to_node, key=key, **data)


def get_edge(from_node, to_node, key=None, is_conn=False, **attr):
    edge_type = EdgeType.CONNECTS if is_conn else EdgeType.DERIVES
    return get_edge_for_type(from_node, to_node, edge_type, key=key, **attr)


def get_edge_for_type(from_node, to_node, edge_type, key=None, **attr):
    # We need a hashable data dict here to be able to use edges in sets
    data = HashableDict(**attr)
    data['type'] = edge_type
    if key is None:
        return from_node, to_node, data
    return from_node, to_node, key, data


def get_edge_type(edge: 'EdgeTuple', default: Optional[EdgeType] = None) -> EdgeType:
    edge_data = edge[-1]
    # Faster than isinstance check
    if edge_data is None or (edge_data.__class__ != dict and edge_data.__class__ != HashableDict):
        return default

    if 'type' not in edge_data:
        if default is None:
            raise ValueError('Edge data not provided')
        return default
    return edge_data['type']


def iter_edges(graph: nx.MultiDiGraph) -> Iterator['EdgeTuple']:
    for edge in graph.edges(keys=True, data=True):
        # if isinstance(edge[3], dict):
        #     edge = (edge[0], edge[1], edge[2], HashableDict(**edge[3]))
        yield edge


def iter_in_edges(graph: nx.MultiDiGraph, node: 'DSGNode', edge_type: EdgeType = None) -> Iterator['EdgeTuple']:
    for edge in graph.in_edges(node, keys=True, data=True):
        if edge_type is None or get_edge_type(edge) == edge_type:
            # if isinstance(edge[3], dict):
            #     edge = (edge[0], edge[1], edge[2], HashableDict(**edge[3]))
            yield edge


def iter_out_edges(graph: nx.MultiDiGraph, node: 'DSGNode', edge_type: EdgeType = None) -> Iterator['EdgeTuple']:
    for edge in graph.out_edges(node, keys=True, data=True):
        if edge_type is None or get_edge_type(edge) == edge_type:
            # if isinstance(edge[3], dict):
            #     edge = (edge[0], edge[1], edge[2], HashableDict(**edge[3]))
            yield edge


def get_edge_data(edge: 'EdgeTuple') -> HashableDict:
    # if isinstance(edge[3], dict):
    #     return HashableDict(**edge[3])
    return edge[-1]



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.encoding import *

__all__ = ['ClosestImputer']


class ClosestImputer(EagerImputer):
    """Imputes by looking for the closest valid design vector based on euclidean or Mahattan distance."""

    def __init__(self, euclidean=True):
        super().__init__()
        self.euclidean = euclidean

    def impute(self, vector: DesignVector, existence: NodeExistence, matrix_mask: MatrixSelectMask) \
            -> Tuple[DesignVector, np.ndarray]:
        design_vectors = self._get_design_vectors(existence)[matrix_mask, :]

        if len(design_vectors) == 0:
            return vector, np.zeros((0, 0), dtype=int)

        elements, target = design_vectors, np.array(vector)
        if self.euclidean:
            dist = self._calc_dist_euclidean(elements, target)
        else:
            dist = self._calc_dist_manhattan(elements, target)
        i_min_dist = int(np.argmin(dist))

        i_mask = np.where(matrix_mask)[0]
        return self._return_imputation(i_mask[i_min_dist], existence)

    @staticmethod
    def _calc_dist_euclidean(elements: np.ndarray, target: np.ndarray) -> np.ndarray:
        return np.sqrt(np.sum((elements-target)**2, axis=1))

    @staticmethod
    def _calc_dist_manhattan(elements: np.ndarray, target: np.ndarray) -> np.ndarray:
        return np.sum(np.abs(elements-target), axis=1)

    def __repr__(self):
        return f'{self.__class__.__name__}(euclidean={self.euclidean})'

    def __str__(self):
        return f'Closest {"Euc" if self.euclidean else "Mht"} Imp'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import math
import itertools
from typing import *
from adsg_core.graph.adsg_basic import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.optimization.evaluator import DSGEvaluator

__all__ = ['GNCEvaluator']


class GNCInstanceNode(NamedNode):
    """Custom node that represents the instance of some object"""

    def __init__(self, base_name: str, idx: int):
        self.idx = idx
        super().__init__(base_name)

    def get_export_title(self) -> str:
        return f'{self.name}[{self.idx}]'

    def __str__(self):
        return f'[{self.name} {self.idx}]'


class GNCTypeNode(NamedNode):
    """Custom node that represents the type of some object"""

    def __init__(self, name: str, type_: str):
        self.type = type_
        super().__init__(name)

    def get_export_title(self) -> str:
        return f'{self.name}: {self.type}'

    def __str__(self):
        return f'[{self.type}]'


class GNCEvaluator(DSGEvaluator):
    """
    Class representing the GN&C (Guidance, Navigation and Control) problem presented in (Chapter 15):
    Crawley 2015: System Architecture - Strategy and Product Development for Complex Systems

    Component mass and probabilities are estimated, as they are not given in the text.

    The problem consists of the connection of a set of sensors to a set of flight computers:
    - System-level mass (minimization) and failure rate (maximization) are optimized
    - The number of sensors and computers can be chosen (1, 2, or 3) --> more reduces failure rate but increases mass
    - The types of sensors and computers can be chosen (A, B, or C) --> each has a different mass/failure rate trade-off
    - Connections between sensors and computers are established
      - Any sensor can connect to any computer
      - Each sensor and computer need at least one connection (otherwise their existence is useless)
    - Sensor/computer type selections are constrained to be non-repeatable
      - E.g. sensor types can be AA, AB, AC, BB, BC, CC (no permutations of these)
      - This is to prevent including isomorphic architectures, architectures where sensors and/or computers, including
        their connections, are permutations of each other and therefor have the same objective values
    """

    mass = {
        'C': {'A': 5., 'B': 10., 'C': 15.},
        'S': {'A': 1., 'B': 2., 'C': 6.},
    }
    failure_rate = {
        'C': {'A': .004, 'B': .0025, 'C': .001},
        'S': {'A': .001, 'B': .0005, 'C': .0001},
    }

    def __init__(self, objective: int = None):
        super().__init__(self.get_adsg(objective=objective))

    @staticmethod
    def get_adsg(objective: int = None):
        metric_nodes = []
        if objective is None or objective == 0:
            metric_nodes.append(MetricNode('mass', direction=-1, type_=MetricType.OBJECTIVE))
        if objective is None or objective == 1:
            metric_nodes.append(MetricNode('failureRate', direction=1, type_=MetricType.OBJECTIVE))
        if len(metric_nodes) == 0:
            raise ValueError('No objectives specified!')

        adsg = BasicDSG()

        # Add top-level node and metrics
        gnc = NamedNode('GNC')
        adsg.add_edges([(gnc, mn) for mn in metric_nodes])

        # Add sensor choices
        sensor = NamedNode('Sensor')
        adsg.add_edge(gnc, sensor)

        sensor_inst_nodes = [GNCInstanceNode('Inst', i) for i in range(3)]
        adsg.add_selection_choice('S', sensor, sensor_inst_nodes)

        sensor_type_nodes = [GNCTypeNode('Type', type_) for type_ in ['A', 'B', 'C']]

        sensor_type_choices = []
        sensor_conn_nodes = []
        for i, si_node in enumerate(sensor_inst_nodes):
            # Type selection
            sensor_type_choices.append(adsg.add_selection_choice(f'ST{i}', si_node, sensor_type_nodes))

            # Connection node, require at least one connection
            conn_node = ConnectorNode(f'SC{i}', deg_spec='+')
            sensor_conn_nodes.append(conn_node)
            adsg.add_edge(si_node, conn_node)

            if i > 0:  # Select previous instances
                adsg.add_edge(si_node, sensor_inst_nodes[i-1])

        # Add computer choices
        computer = NamedNode('Comp')
        adsg.add_edge(gnc, computer)

        comp_inst_nodes = [GNCInstanceNode('Inst', i) for i in range(3)]
        adsg.add_selection_choice('C', computer, comp_inst_nodes)

        comp_type_nodes = [GNCTypeNode('Type', type_) for type_ in ['A', 'B', 'C']]

        comp_type_choices = []
        comp_conn_nodes = []
        for i, ci_node in enumerate(comp_inst_nodes):
            # Type selection
            comp_type_choices.append(adsg.add_selection_choice(f'CT{i}', ci_node, comp_type_nodes))

            # Connection node, require at least one connection
            conn_node = ConnectorNode(f'CC{i}', deg_spec='+')
            comp_conn_nodes.append(conn_node)
            adsg.add_edge(ci_node, conn_node)

            if i > 0:  # Select previous instances
                adsg.add_edge(ci_node, comp_inst_nodes[i-1])

        # Add connection choice
        adsg.add_connection_choice('Conn', sensor_conn_nodes, comp_conn_nodes)

        # Set start node
        adsg = adsg.set_start_nodes({gnc})

        # Constraint sensor/computer type selections
        adsg = adsg.constrain_choices(ChoiceConstraintType.UNORDERED, comp_type_choices)
        adsg = adsg.constrain_choices(ChoiceConstraintType.UNORDERED, sensor_type_choices)

        return adsg

    def _evaluate(self, dsg: DSGType, metric_nodes: List[MetricNode]) -> Dict[MetricNode, float]:
        # Analyze the architecture
        sensors, computers, conns = self._analyze_arch(dsg)

        # Calculate metrics: the outputs of this function should be a dict mapping requested metric nodes to values
        results = {}
        for metric_node in metric_nodes:
            if metric_node.name == 'mass':
                results[metric_node] = self.calc_mass(sensors, computers)
            elif metric_node.name == 'failureRate':
                results[metric_node] = self.calc_failure_rate(sensors, computers, conns)

        return results

    @staticmethod
    def _analyze_arch(adsg: BasicDSG) -> Tuple[list, list, list]:

        def _analyze_object(object_root_node):
            obj_types = []
            connector_nodes = {}

            # Find object instance nodes
            inst_nodes = []
            for obj_inst_node in adsg.derived_nodes(object_root_node):
                if isinstance(obj_inst_node, GNCInstanceNode):
                    inst_nodes.append(obj_inst_node)

            # Loop over sorted instance nodes
            for i, obj_inst_node in enumerate(sorted(inst_nodes, key=lambda n: n.idx)):
                for next_node in adsg.next(obj_inst_node):
                    # Record selected type
                    if isinstance(next_node, GNCTypeNode):
                        obj_types.append(next_node.type)

                    # Record connector nodes
                    elif isinstance(next_node, ConnectorNode):
                        connector_nodes[next_node] = i

            return obj_types, connector_nodes

        # Analyze object types
        sensors, sensor_conns = [], {}
        computers, comp_conns = [], {}
        for node in adsg.graph.nodes:
            if isinstance(node, NamedNode):
                if node.name == 'Sensor':
                    sensors, sensor_conns = _analyze_object(node)
                elif node.name == 'Comp':
                    computers, comp_conns = _analyze_object(node)

        # Get object connections
        conns = []
        for src_node, src_idx in sensor_conns.items():
            for tgt_node in adsg.next(src_node, edge_type=EdgeType.CONNECTS):
                tgt_idx = comp_conns[tgt_node]
                conns.append((src_idx, tgt_idx))

        return sensors, computers, conns

    @classmethod
    def calc_mass(cls, sensors, computers):
        sensor_mass = sum([cls.mass['S'][type_] for type_ in sensors])
        computer_mass = sum([cls.mass['C'][type_] for type_ in computers])
        return sensor_mass+computer_mass

    @classmethod
    def calc_failure_rate(cls, sensors, computers, conns):

        def system_fails(sensor_failed, computer_failed):
            # Find remaining connections for the failed sensors and computers
            remaining_conns = [conn for conn in conns if not sensor_failed[conn[0]] and not computer_failed[conn[1]]]

            # If there are no remaining connections, the system has failed
            return len(remaining_conns) == 0

        # Get item failure rates
        rate = cls.failure_rate
        failure_rates = [rate['S'][type_] for type_ in sensors]+[rate['C'][type_] for type_ in computers]

        # Loop over number of failures
        failure_rate = 0.
        n_s = len(sensors)
        n_c = len(computers)
        n_obj = n_s+n_c
        for n_failed in range(n_obj):
            base_failures = [i < n_failed+1 for i in range(n_obj)]

            # Iterate over possible failure permutations
            for failure_scheme in set(itertools.permutations(base_failures)):
                s_failed = failure_scheme[:n_s]
                c_failed = failure_scheme[n_s:]

                # Check if system fails
                if system_fails(s_failed, c_failed):
                    # Determine probability of this state happening
                    prob = 1.
                    for i, f_rate in enumerate(failure_rates):
                        if failure_scheme[i]:
                            prob *= f_rate

                    failure_rate += prob

        return -math.log10(failure_rate)


if __name__ == '__main__':
    # After export, visualize contents using https://viz-js.com/
    GNCEvaluator.get_adsg().export_dot('gnc.dot')

    evaluator = GNCEvaluator()
    for _ in range(10):
        graph, _, _ = evaluator.get_graph(evaluator.get_random_design_vector())
        evaluator.evaluate(graph)
